/**
 * Copyright (C) 2003-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/**
 * Copyright (C) 2003-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.framework.core;

import java.util.*;

/**
 * This class represents the &lt;Sequence&gt; tag as defined by the SyncML
 * representation specifications.
 * A sequence can contain the following commands: Add, Replace, Delete, Copy,
 * Atomic, Map, Move, Sync, Get, Alert, Exec.
 *
 * @version $Id: Sequence.java,v 1.5 2007/01/11 11:48:50 nichele Exp $
 */
public class Sequence
extends AbstractCommand
implements java.io.Serializable {

    // --------------------------------------------------------------- Constants

    public static String COMMAND_NAME = "Sequence";

    // ------------------------------------------------------------ Private data

    private ArrayList commands = new ArrayList();

    // ------------------------------------------------------------ Constructors

    /**
     * For serialization purposes
     */
    protected Sequence() {}

    /**
     * Create a new Sequence object. The commands in <i>commands</i>
     * must be of the allowed types.
     *
     * @param cmdID command identifier - NOT NULL
     * @param noResp is &lt;NoREsponse/&gt; required?
     * @param meta meta information
     * @param commands the sequenced commands - NOT NULL
     *
     * @throws java.lang.IllegalArgumentException is any of the parameters is invalid
     *
     */
    public Sequence(
               final CmdID             cmdID     ,
               final boolean           noResp    ,
               final Meta              meta      ,
               final AbstractCommand[] commands  ) {
        super(cmdID, noResp);

        setMeta(meta);
        setCommands(commands);
    }

    // ---------------------------------------------------------- Public methods

    /**
     * Gets an array of AbstractCommand
     *
     * @return an array of command objects
     */
    public ArrayList getCommands() {
        return this.commands;
    }

    /**
     * Sets the sequenced commands. The given commands must be of the allowed
     * types.
     *
     * @param commands the commands - NOT NULL and o the allawed types
     *
     * @throws IllegalArgumentException if the constraints are not met
     */
    public void setCommands(AbstractCommand[] commands) {
        if (commands == null) {
            throw new IllegalArgumentException("commands cannot be null");
        }

        for (int i = 0; i < commands.length; i++) {
            if (commands[i] == null) {
                throw new IllegalArgumentException("commands[" + i +"] cannot be null");
            } else if (   (!(commands[i] instanceof Add))
                       && (!(commands[i] instanceof Alert))
                       && (!(commands[i] instanceof Atomic))
                       && (!(commands[i] instanceof Copy))
                       && (!(commands[i] instanceof Delete))
                       && (!(commands[i] instanceof Exec))
                       && (!(commands[i] instanceof Get))
                       && (!(commands[i] instanceof Map))
                       && (!(commands[i] instanceof Move))
                       && (!(commands[i] instanceof Replace))
                       && (!(commands[i] instanceof Sync))) {
                throw new IllegalArgumentException(
                    "commands[" + i + "] cannot be a " + commands[i].getName()
                );
            }
        }
        this.commands.clear();
        this.commands.addAll(Arrays.asList(commands));
    }

    /**
     * Returns the command name
     *
     * @return the command name
     */
    public String getName() {
        return Sequence.COMMAND_NAME;
    }
}
