/**
 * Copyright (C) 2003-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.framework.engine.source;

import java.util.Collection;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import java.sql.Timestamp;

import com.funambol.framework.engine.SyncProperty;
import com.funambol.framework.engine.SyncItemKey;
import com.funambol.framework.engine.SyncItem;
import com.funambol.framework.engine.SyncItemImpl;
import com.funambol.framework.engine.SyncItemState;

/*
 * This source represents a <i>SyncSource</i> where all items are kept in memory.
 * A <i>MemorySyncSource</i> can be created passing in the constructor the name
 * of the source and the items or creating an anonymous source with the default
 * constructor and then initialize the source with <i>initialize(...)</i>.
 * <p>
 * <code>existingSyncItems</code> represents the items on the device.
 * <p>
 * <code>newSyncItems</code>, <code>updatedSyncItems</code> and
 * <code>deletedSyncItems</code> are insted of the modifications reported with
 * the last message.
 *
 *
 * @version $Id: MemorySyncSource.java,v 1.10 2007/03/05 14:54:27 nichele Exp $
 */
public class MemorySyncSource
extends AbstractSyncSource
implements SyncSource, java.io.Serializable {

    // --------------------------------------------------------------- Constants

    public static final String NAME = "MemorySyncSource";

    // ----------------------------------------------------- Private data member

    // the content of the SyncML client
    private List existingSyncItems;
    private List newSyncItems     ;
    private List updatedSyncItems ;
    private List deletedSyncItems ;

    private List allNewSyncItems     ;
    private List allUpdatedSyncItems ;
    private List allDeletedSyncItems ;

    /**
     * LO SyncItem - not null if any LO is being retrieved
     */
    private SyncItem lo;

    // ------------------------------------------------------------ Constructors

    public MemorySyncSource() {
        lo = null;
    }

    public MemorySyncSource(String name, String sourceURI) {
        super(name, sourceURI);
        lo = null;
        initialize(new ArrayList(), new ArrayList(), new ArrayList());

        allNewSyncItems     = new ArrayList();
        allUpdatedSyncItems = new ArrayList();
        allDeletedSyncItems = new ArrayList();

        setExistingItems(new ArrayList());
    }

    public MemorySyncSource(String name            ,
                            List unchangedSyncItems,
                            List deletedSyncItems  ,
                            List newSyncItems      ) {
      //
      // The items are added so no need to clone the lists
      //
      super(name);
      lo = null;
      initialize(deletedSyncItems, newSyncItems, updatedSyncItems);
      setExistingItems(new ArrayList());
    }

    // ---------------------------------------------------------- Public Methods

    /**
     * Sets the name of the source.
     *
     * @param name the new name
     */
    public void setName(String name) {
        this.name = name;
    }



    /*
     * @see SyncSource
     *
     */
    public SyncItem[] getSyncItemsFromIds(SyncItemKey[] syncItemKeys)
    throws SyncSourceException {
        List ret           = new ArrayList();
        SyncItem  syncItem = null;

        for (int i=0; ((syncItemKeys != null) && (i<syncItemKeys.length)); ++i) {
            syncItem = getSyncItemFromId(syncItemKeys[i]);

            if (syncItem != null) {
                ret.add(syncItem);
            }
        }

        return (SyncItem[])ret.toArray(new SyncItem[ret.size()]);
    }

    /*
     * @see SyncSource
     */
    public SyncItemKey[] getNewSyncItemKeys(Timestamp sinceTs,
                                            Timestamp untilTs)
    throws SyncSourceException {
        return extractKeys(newSyncItems);
    }

    /**
     * Returns the new items
     * @return SyncItem[]
     */
    public SyncItem[] getNewSyncItems() {
        return (SyncItem[])newSyncItems.toArray(new SyncItem[0]);
    }

    /*
     * @see SyncSource
     */
    public SyncItemKey[] getDeletedSyncItemKeys(Timestamp sinceTs,
                                                Timestamp untilTs)
    throws SyncSourceException {
        return extractKeys(deletedSyncItems);
    }

    /**
     * Returns the deleted items
     * @return SyncItem[]
     */
    public SyncItem[] getDeletedSyncItems() {
        return (SyncItem[])deletedSyncItems.toArray(new SyncItem[0]);
    }

    /**
     * @return an array of keys containing the <i>SyncItem</i>'s key of the updated
     *        items after the last synchronizazion.
     */
    public SyncItemKey[] getUpdatedSyncItemKeys(Timestamp sinceTs,
                                                Timestamp untilTs)
    throws SyncSourceException {
        return extractKeys(updatedSyncItems);
    }

    /**
     * Returns the updated items
     * @return SyncItem[]
     */
    public SyncItem[] getUpdatedSyncItems() {
        return (SyncItem[])updatedSyncItems.toArray(new SyncItem[0]);
    }

    /**
     * @see SyncSource
     */
    public SyncItem[] getAllSyncItems()
    throws SyncSourceException {
        return (SyncItem[])existingSyncItems.toArray(new SyncItemImpl[0]);
    }

    /**
     * @see SyncSource
     */
    public SyncItemKey[] getAllSyncItemKeys()
    throws SyncSourceException {
        return extractKeys(existingSyncItems);
    }


    /**
     * @see SyncSource
     */
    public boolean isModified() {

        return (getDeletedSyncItems().length > 0)
                || (getNewSyncItems().length     > 0)
                || (getUpdatedSyncItems().length > 0);

    }

    /**
     * Initializes this MemorySyncSource with items.
     *
     * @param deletedItems the deleted items
     * @param newItems the created items
     * @param updatedItems the updated items
     */
    public void initialize(List deletedItems ,
                           List newItems     ,
                           List updatedItems ) {

       //
       // Large object handling
       // =====================
       // If the server is still retrieving a large object, lo will be set to
       // the item being retrieved and the new piece of data must be merged to
       // the previous content. The partial item must be then removed by the
       // list that contains it. When the LO is fully retrieved, the complete
       // item will be updated in the containing list and from now on it will
       // like any other item. The only difference is that the property
       // PROPERTY_SIZE will be set to the expected content size
       //
       if (deletedSyncItems == null) {
           deletedSyncItems = new ArrayList();
       } else {
           deletedSyncItems.clear();
       }

       if (newSyncItems == null) {
           newSyncItems = new ArrayList();
       } else {
           newSyncItems.clear();
       }

       if (updatedSyncItems == null) {
           updatedSyncItems = new ArrayList();
       } else {
           updatedSyncItems.clear();
       }

       deletedSyncItems.addAll(deletedItems)  ;
       newSyncItems.addAll(newItems)          ;
       updatedSyncItems.addAll(updatedItems ) ;

       boolean cont = false;
       do {
           cont = handleLOs();
       } while (cont);

       //
       // We have to make sure that the complete view is consistent so that
       // new, modified or deleted items are not in the unchanged list; that
       // deleted items are not in new, update and unchanged lists, and so on.
       //
       SyncItem syncItem = null;
       Iterator i = deletedSyncItems.iterator();
       while (i.hasNext()) {
           syncItem = (SyncItem)i.next();
           newSyncItems.remove(syncItem)      ;
           updatedSyncItems.remove(syncItem)  ;
       }

       i = updatedSyncItems.iterator();
       while (i.hasNext()) {
           syncItem = (SyncItem)i.next();
           newSyncItems.remove(syncItem)      ;
           deletedSyncItems.remove(syncItem);
       }

       i = newSyncItems.iterator();
       while (i.hasNext()) {
           syncItem = (SyncItem)i.next();
           updatedSyncItems.remove(syncItem);
           deletedSyncItems.remove(syncItem);
       }
    }


    /**
     * Sets the existing items
     *
     * @param items existing items
     *
     */
    public void setExistingItems(List items) {
        existingSyncItems = items;
    }


    /**
     * Search the items for an item with exactly the same binary content.
     *
     * @return an item from a twin item. Each source implementation is free to
     *         interpret this as it likes (i.e.: comparing all fields).
     *
     * @param principal
     * @param syncItemTwin the twin item
     *
     * @throws SyncSourceException in case of errors (for instance if the
     *         underlying data store runs into problems)
     *
     */
    public SyncItemKey[] getSyncItemKeysFromTwin(SyncItem syncItemTwin)
    throws SyncSourceException {

        String contentToSearch = null;

        contentToSearch = new String(syncItemTwin.getContent());

        SyncItem[] all = getAllSyncItems();

        String content    = null;
        List itemKeys     = new ArrayList();

        for (int i = 0; i < all.length; ++i) {
            content = new String(all[i].getContent());
            if (content.equals(contentToSearch)) {
                itemKeys.add(all[i].getKey());
            }
        }

        return (SyncItemKey[])itemKeys.toArray(new SyncItemKey[itemKeys.size()]);
    }

    /**
     * Two items are considered twin if their binary content, converted to
     * String are equal.
     *
     * @param principal
     * @param syncItemTwins the twin items
     *
     * @return an array of items from corresponding twin items.
     *
     * @throws SyncSourceException in case of errors (for instance if the
     *         underlying data store runs into problems)
     *
     */
    public SyncItem[] getSyncItemsFromTwins(SyncItem[] syncItemTwins)
    throws SyncSourceException {
        ArrayList ret = new ArrayList();

        String[] contents = new String[syncItemTwins.length];

        for (int i=0; i<syncItemTwins.length; ++i) {
            contents[i] = new String(syncItemTwins[i].getContent());
        }

        SyncItem[] all = getAllSyncItems();

        String content = null;

        for (int i=0; i<all.length; ++i) {
            content = new String(all[i].getContent());
            for (int j=0; ((content != null) && (j<contents.length)); ++j) {
                if (content.equals(contents[j])) {
                    ret.add(all[i]);
                    break;
                }
            }
        }

        return (SyncItem[])ret.toArray(new SyncItem[ret.size()]);
    }

    /**
     * see SyncSource
     */
    public boolean isSyncItemInFilterClause(SyncItem item) throws SyncSourceException {
        return true;
    }

    /**
     * see SyncSource
     */
    public boolean isSyncItemInFilterClause(SyncItemKey key) throws SyncSourceException {
        return true;
    }


    // --------------------------------------------------------- Private methods

    private SyncItemKey[] extractKeys(Collection syncItems) {
        SyncItemKey[] keys = new SyncItemKey[syncItems.size()];

        SyncItem syncItem = null;
        int j = 0;
        for(Iterator i = syncItems.iterator(); i.hasNext(); ++j) {
           syncItem = (SyncItem)i.next();

           keys[j] = syncItem.getKey();
        } // next i, j

        return keys;
    }

    private SyncItemKey[] extractKeys(SyncItem[] syncItems) {
        SyncItemKey[] keys = new SyncItemKey[syncItems.length];

        for(int j = 0; j<syncItems.length; ++j) {
           keys[j] = syncItems[j].getKey();
        } // next i, j

        return keys;
    }

    public void commitSync() throws SyncSourceException {
        allNewSyncItems.addAll(newSyncItems);
        allUpdatedSyncItems.addAll(updatedSyncItems);
        allDeletedSyncItems.addAll(deletedSyncItems);
    }

    public SyncItem[] getAllNewSyncItems()
    throws SyncSourceException {
        return (SyncItem[])allNewSyncItems.toArray(new SyncItem[0]);
    }

    public SyncItem[] getAllUpdatedSyncItems()
    throws SyncSourceException {
        return (SyncItem[])allUpdatedSyncItems.toArray(new SyncItem[0]);
    }

    public SyncItem[] getAllDeletedSyncItems()
    throws SyncSourceException {
        return (SyncItem[])allDeletedSyncItems.toArray(new SyncItem[0]);
    }

    // --------------------------------------------------------- Private Methods
    private SyncItem setSyncItem(SyncItem syncItem)
        throws SyncSourceException {

        newSyncItems.remove(syncItem);
        deletedSyncItems.add(syncItem);
        updatedSyncItems.add(syncItem);

        SyncItemKey mappedKey = ((SyncItemImpl)syncItem).getMappedKey();
        String mappedKeyValue = (mappedKey == null) ? null
                                : mappedKey.getKeyAsString();

        SyncItem newSyncItem = new SyncItemImpl(
            this,
            syncItem.getKey().getKeyAsString(),
            mappedKeyValue,
            SyncItemState.NEW,
            syncItem.getContent(),
            syncItem.getFormat(),
            syncItem.getType(),
            syncItem.getTimestamp()
        );

        return newSyncItem;
    }

    /**
     * Look for LO and handle it accordingly.
     *
     * @return true if there could be something still to process, false otherwise
     */
    private boolean handleLOs() {
        boolean ret = false;

        SyncItem chunk = null;
        int p = -1;

        if (lo != null) {
            // We are reading a LO
            p = newSyncItems.indexOf(lo);
            if (p >= 0) {
                chunk = (SyncItem)newSyncItems.get(p);
            } else {
                p = updatedSyncItems.indexOf(lo);
                if (p >= 0) {
                    chunk = (SyncItem)updatedSyncItems.get(p);
                }
            }

            if (chunk != null) {
                mergeLOChunk(chunk);

                //
                // if the chunk is still in the state 'P', it means that the
                // current LO is not complete. Therefore, we need to not add
                // the item to the news or updated lists.
                //
                if (chunk.getState() != SyncItemState.PARTIAL) {
                switch (lo.getState()) {
                    case SyncItemState.PARTIAL:
                        //
                        // just a new chunk
                        //
                        break;

                    case SyncItemState.NEW:
                        //
                        // Last chunk of an addition
                        //
                        newSyncItems.remove(chunk);
                        newSyncItems.add(lo);
                            lo = null; ret = true;
                        break;

                    case SyncItemState.UPDATED:
                    case SyncItemState.SYNCHRONIZED:
                        //
                        // Last chunk of an update
                        //
                        updatedSyncItems.remove(chunk);
                        updatedSyncItems.add(lo);
                            lo = null; ret = true;
                        break;
                }
                } // chunk state
            } // chunk != null
        } else {
            //
            // Look for the first chunk
            //

            Iterator i = newSyncItems.iterator();
            while(i.hasNext()) {
                chunk = (SyncItem)i.next();

                if (chunk.getState() == SyncItemState.PARTIAL) {
                    lo = newSyncItem(chunk);
                    lo.setState(SyncItemState.NEW);
                    break;
                }
            }

            i = updatedSyncItems.iterator();
            while(i.hasNext()) {
                chunk = (SyncItem)i.next();

                if (chunk.getState() == SyncItemState.PARTIAL) {
                    lo = newSyncItem(chunk);
                    lo.setState(SyncItemState.UPDATED);
                    break;
                }
            }
        }

        return ret;
    }

    /**
     * Merges the given <code>chunk</code> with the existing <code>lo</code>
     * item.
     *
     * @param chunk the chunk to merge
     *
     */
    private void mergeLOChunk(SyncItem chunk) {
        if (lo == null) {
            //
            // we should never get here!
            //
            return;
        }

        byte[] loData    = lo.getContent();
        byte[] chunkData = chunk.getContent();

        SyncProperty oldSize =
            ((SyncItemImpl)lo).getProperty(SyncItemImpl.PROPERTY_DECLARED_SIZE);

        //
        // The chunk has the DECLARED_SIZE property set: using it has real size
        // of item (probably some synclets have modified the content of the large
        // object item and therefore have modified its size)
        //
        SyncProperty newSize =
            ((SyncItemImpl)chunk).getProperty(SyncItemImpl.PROPERTY_DECLARED_SIZE);

        byte[] allData = null;

        if (newSize != null) {
            oldSize = newSize;
            //
            // If the new size is not null, it means the item contains all data
            // and not just a chunk. This is not expected by the protocol but it's
            // needed to handle the large object in the synclets.
            //
            allData = new byte[chunkData.length];
            allData = chunkData;
        } else {
            allData = new byte[loData.length + chunkData.length];

            System.arraycopy(loData, 0, allData, 0, loData.length);
            System.arraycopy(chunkData, 0, allData, loData.length, chunkData.length);
        }

        lo.setContent  (allData);
        lo.setFormat   (chunk.getFormat());
        lo.setType     (chunk.getType());
        lo.setTimestamp(chunk.getTimestamp());
        ((SyncItemImpl)lo).setProperties(((SyncItemImpl)chunk).getProperties());
        ((SyncItemImpl)lo).setProperty(oldSize);
        ((SyncItemImpl)lo).setProperty(
                new SyncProperty(SyncItemImpl.PROPERTY_CONTENT_SIZE,
                                 new Long(allData.length))
             );
    }

    /**
     * Create a new <i>SyncItem</i> with the content of an existing one.
     *
     * @param key
     * @param fromSyncItem
     */
    private SyncItem newSyncItem(SyncItem fromSyncItem) {
        SyncItemKey mappedKey = ((SyncItemImpl)fromSyncItem).getMappedKey();

        String mappedKeyValue = (mappedKey == null)
                              ? null
                              : mappedKey.getKeyAsString()
                              ;

        SyncItemKey parentKey = ((SyncItemImpl)fromSyncItem).getParentKey();

        String parentKeyValue = (parentKey == null)
                              ? null
                              : parentKey.getKeyAsString()
                              ;

        SyncItemImpl syncItem
            = new SyncItemImpl(fromSyncItem.getSyncSource()                ,
                               fromSyncItem.getKey().getKeyAsString()      ,
                               parentKeyValue                              ,
                               mappedKeyValue                              ,
                               fromSyncItem.getState()                     ,
                               fromSyncItem.getContent()                   ,
                               fromSyncItem.getFormat()                    ,
                               fromSyncItem.getType()                      ,
                               fromSyncItem.getTimestamp()
              );

        if (syncItem     instanceof SyncItemImpl &&
            fromSyncItem instanceof SyncItemImpl) {

            ((SyncItemImpl)syncItem).setProperties(
               ((SyncItemImpl)fromSyncItem).getProperties()
            );

        }


        return syncItem;
    }

    /**
     * Updates the given item
     * @param syncItem SyncItem
     * @return SyncItem
     * @throws SyncSourceException
     */
    public SyncItem updateSyncItem(SyncItem syncItem)
    throws SyncSourceException {
        return setSyncItem(syncItem);
    }


    /**
     * Adds the given item
     * @param syncItem SyncItem
     * @return SyncItem
     * @throws SyncSourceException
     */
    public SyncItem addSyncItem(SyncItem syncItem)
    throws SyncSourceException {
        return setSyncItem(syncItem);
    }

    /**
     * Removes the given item
     * @param syncItem SyncItem
     * @param time Timestamp
     * @throws SyncSourceException
     */
    public void removeSyncItem(SyncItemKey syncItemKey, Timestamp time)
    throws SyncSourceException {
        if (syncItemKey == null) {
            return;
        }
        //
        // We create a dummyItem just to remove the item with the same key
        SyncItem dummyItem = new SyncItemImpl(this, syncItemKey.getKeyAsString());

        newSyncItems.remove      (dummyItem);
        updatedSyncItems.remove  (dummyItem);
        deletedSyncItems.add     (dummyItem);
    }

    /**
     * Removes the given item
     * @param syncItem SyncItem
     * @param time Timestamp
     * @param softDelete is a soft delete
     * @throws SyncSourceException
     */
    public void removeSyncItem(SyncItemKey syncItemKey,
                               Timestamp   time,
                               boolean     softDelete)
    throws SyncSourceException {

        //
        // Here, we handle all deletions as hard deletion
        //
        removeSyncItem(syncItemKey, time);
    }

    /*
     * @see SyncSource
     *
     * This implementation cycles through all SyncItems looking for the
     * specified key.
     */
    public SyncItem getSyncItemFromId(SyncItemKey syncItemKey)
    throws SyncSourceException {
        SyncItem[] all = this.getAllSyncItems();

        for (int i=0; ((all != null) && (i<all.length)); ++i) {
            if (syncItemKey.equals(all[i].getKey())) {
                return all[i];
            }
        }

        return null;
    }

    /**
     * @see SyncSource
     */
    public char getSyncItemStateFromId(SyncItemKey syncItemKey)
    throws SyncSourceException {
        SyncItem[] all = this.getAllSyncItems();

        for (int i = 0; ( (all != null) && (i < all.length)); ++i) {
            if (syncItemKey.equals(all[i].getKey())) {
                return all[i].getState();
            }
        }

        return SyncItemState.NOT_EXISTING;
    }

    /**
     * @see SyncSource
     */
    public void setOperationStatus(String        operationName,
                                   int           status,
                                   SyncItemKey[] keys) {

    }

}
