/**
 * Copyright (C) 2006-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */

package com.funambol.framework.logging;

import java.util.Hashtable;
import java.util.Set;

import org.apache.log4j.MDC;

/**
 * It's the repository of the information used by the Logging architecture to store
 * thread variables. Uses org.apache.log4j.MDC as real implementation of the
 * repository, providing just some useful methods to hidden the MDC use.
 * <br/>
 * See log4j documentation for detail about MDC class.
 *
 * @version $Id: LogContext.java,v 1.5 2007/02/14 12:39:33 nichele Exp $
 */
public class LogContext {

    // --------------------------------------------------------------- Constants
    private static final String PARAMETER_USER_NAME  = "userName" ;
    private static final String PARAMETER_SESSION_ID = "sessionId";
    private static final String PARAMETER_DEVICE_ID  = "deviceId" ;
    private static final String PARAMETER_SOURCE_URI = "sourceURI";
    private static final String PARAMETER_THREAD_ID  = "threadId" ;

    // ------------------------------------------------------------- Constructor

    // ---------------------------------------------------------- Public methods

    /**
     * Returns the userName stored in the thread space
     * @return the userName stored in the thread space
     */
    public static String getUserName() {
        return (String)MDC.get(PARAMETER_USER_NAME);
    }

    /**
     * Stores in the thread space the given userName.
     * If the given userName is null, the userName in the thread space is removed
     * @param userName the userName to set
     */
    public static void setUserName(String userName) {
        if (userName == null) {
            MDC.remove(PARAMETER_USER_NAME);
        } else {
            MDC.put(PARAMETER_USER_NAME, userName);
        }
    }

    /**
     * Returns the thread id stored in the thread space
     * @return the thread id stored in the thread space
     */
    public static String getThreadId() {
        return (String)MDC.get(PARAMETER_THREAD_ID);
    }

    /**
     * Stores in the thread space the given thread id.
     * If the given thread id is null, the thread id in the thread space is removed
     * @param threadId the thread id to set
     */
    public static void setThreadId(String threadId) {
        if (threadId == null) {
            MDC.remove(PARAMETER_THREAD_ID);
        } else {
            MDC.put(PARAMETER_THREAD_ID, threadId);
        }
    }


    /**
     * Returns the device id stored in the thread space
     * @return the device id stored in the thread space
     */
    public static String getDeviceId() {
        return (String)MDC.get(PARAMETER_DEVICE_ID);
    }

    /**
     * Stores in the thread space the given deviceId.
     * If the given deviceId is null, the deviceId in the thread space is removed.
     * @param deviceId the device id to set
     */
    public static void setDeviceId(String deviceId) {
        if (deviceId == null) {
            MDC.remove(PARAMETER_DEVICE_ID);
        } else {
            MDC.put(PARAMETER_DEVICE_ID, deviceId);
        }
    }

    /**
     * Returns the session id stored in the thread space
     * @return the session id stored in the thread space
     */
    public static String getSessionId() {
        return (String)MDC.get(PARAMETER_SESSION_ID);
    }

    /**
     * Stores in the thread space the given sessionId.
     * If the given sessionId is null, the sessionId in the thread space is removed.
     * @param sessionId the session id to set
     */
    public static void setSessionId(String sessionId) {
        if (sessionId == null) {
            MDC.remove(PARAMETER_SESSION_ID);
        } else {
            MDC.put(PARAMETER_SESSION_ID, sessionId);
        }
    }

    /**
     * Returns the source uri stored in the thread space
     * @return the source uri stored in the thread space
     */
    public static String getSourceURI() {
        return (String)MDC.get(PARAMETER_SOURCE_URI);
    }

    /**
     * Stores in the thread space the given sourceURI.
     * If the given sourceURI is null, the sourceURI in the thread space is removed.
     * @param sourceURI the source uri to set
     */
    public static void setSourceURI(String sourceURI) {
        if (sourceURI == null) {
            MDC.remove(PARAMETER_SOURCE_URI);
        } else {
            MDC.put(PARAMETER_SOURCE_URI, sourceURI);
        }
    }

    /**
     * Stores in the thread space the given value with the given key.
     * If the given value is null, the given key in the thread space is removed.
     * @param key the key of the object to store
     * @param value the object to store
     */
    public static void set(String key, Object value) {
        if (value == null) {
            MDC.remove(key);
        } else {
            MDC.put(key, value);
        }
    }

    /**
     * Removes from the thread space all values
     */
    public static void clear() {
        Hashtable context = MDC.getContext();
        if (context != null) {
            context.clear();
        }

    }

    /**
     * Returns the value of the variable in the thread space with the given key
     * @param key the key of the value
     * @return the value of the given key
     */
    public static Object getValue(String key) {
        return MDC.get(key);
    }

    /**
     * Returns all values in the thread space
     * @return all values in the thread space
     */
    public static Hashtable getValues() {
        return MDC.getContext();
    }

    /**
     * Sets in the thread space all values contained in the given Hashtable deleting
     * the old ones (a <code>clear()</code> is performed before to set the values)
     * @param values the values to set in the thread space
     */
    public static void setValues(Hashtable values) {
        clear();
        if (values == null) {
            return;
        }
        Set keys = values.keySet();
        for (Object key : keys) {
            MDC.put((String)key, values.get(key));
        }
    }

    // --------------------------------------------------------- Private methods


}
