/**
 * Copyright (C) 2003-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */

package com.funambol.framework.protocol;

import com.funambol.framework.protocol.IdGenerator;
import com.funambol.framework.protocol.SimpleIdGenerator;
import com.funambol.framework.core.CmdID;

/**
 * This class creates command ids. Note that when a new CommandIdentifier is
 * created, it is stored also in <i>currentId</i>, which is returned by
 * <i>current()</i>. This is to avoid that <i>current()</i> creates a new
 * CommandIdentifier object each time is called.
 * <p>
 * This class is thread-safe.
 *
 *
 * @version $Id: CommandIdGenerator.java,v 1.3 2007/01/11 11:48:50 nichele Exp $
 */
public class CommandIdGenerator
        implements java.io.Serializable {

  // ---------------------------------------------------------- Private fields

  /**
   * The underlying IdGenerator object
   */
  private IdGenerator idGenerator = null;

  /**
   * The current identifier
   */
  private CmdID currentId = null;

  // ------------------------------------------------------------ Constructors

  /** Creates a new instance of CommandIdGenerator */
  public CommandIdGenerator() {
    this(new SimpleIdGenerator());

  }

  /**
   * Creates a new instance of CommandIdGenerator given an IdGenerator
   *
   * @param idGenerator the idGenerator object - NOT NULL
   */
  public CommandIdGenerator(IdGenerator idGenerator) {
    if (idGenerator == null) {
      throw new NullPointerException("idGenerator cannot be null!");
    }
    this.idGenerator = idGenerator;
  }

  // ---------------------------------------------------------- Public methods

  /**
   * Returns a new generated command id.
   *
   * @return a new generated command id.
   */
  public synchronized CmdID next() {
    return (currentId = new CmdID(idGenerator.next()));
  }

  /**
   * Reset the Id counter
   */
  public synchronized void reset() {
    idGenerator.reset();
  }

  /**
   * Returns the last generated command id (which is the current command id).
   *
   * @return the last generated command id
   */
  public synchronized CmdID current() {
    return currentId;
  }
}
