/**
 * Copyright (C) 2005-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */

package com.funambol.framework.server.inventory;

import com.funambol.framework.filter.Clause;
import com.funambol.framework.server.*;

/**
 * The interface to store and read capabilities and devices.
 *
 * @version $Id: DeviceInventory.java,v 1.4 2007/01/11 11:48:59 nichele Exp $
 */
public interface DeviceInventory
{
    /* Searches the capabilities identified:
     *  -id
     *  or, if the id is null:
     *  - by the set:Model,Manufacturer,Firmware version,Software version(if all these properties are not empty.Otherwise the method has to return false)
     *  and fills the given object with the capabilities in the inventory. Returns true if the capabilities were found, false otherwise.
     * @param caps Capabilities to get
     * @return true if capabilities were found, false otherwise
     * @throws DeviceInventoryException
     */
    public boolean getCapabilities(Capabilities caps) throws DeviceInventoryException;

    /**
     *  Add or update the given capabilities to the inventory.If caps.getId() is null, the capabilities have to be added to the inventory.
     *  If caps.getId() isn't null but in the inventory there aren't the given capabilities an add will be performed.
     *  If caps.getId() isn't null and in the inventory there are the given capabilities an update will be performed.
     *  If the capabilities are added and the concrete implementation creates a new id for the capabilities, that id is caps.setId() must be called in order to update it.
     * @param caps capabilities object to be saved
     * @throws DeviceInventoryException
     */
    public void setCapabilities(Capabilities  caps) throws DeviceInventoryException;

    /**
     * Retrieves the capabilities that match the given clause filter.
     * @param clause the clause capabilities to match
     * @return array of Cpabilities that match the given clause
     * @throws DeviceInventoryException
     */
    public Capabilities[] queryCapabilities(Clause clause) throws DeviceInventoryException;

    /**
     * Returns the number of the capabilities records in the repository that match the given clause filter.
     * If clause is null or an instance of AllClause, count all records.
     * @param clause the clause capabilities to match
     * @return number of capabilities that match the given clause
     * @throws DeviceInventoryException
     */
    public int countCapabilities(Clause clause) throws DeviceInventoryException;

    /**
     * Searches the device identified by device.getDeviceId() and fills the given object with the info in the inventory
     *
     * @param device the object to be filled with informations from repository ( deviceId must be set)
     * @return true if the device was found, false otherwise
     * @throws DeviceInventoryException
     */
    public boolean getDevice(Sync4jDevice device) throws DeviceInventoryException;

    /**
     * Searches the device identified by device.getDeviceId() and fills the given object with the info in the inventory
     * specifying whether device capabilities should be retrieved as well.
     *
     * @param device the object to be filled with informations from repository ( deviceId must be set)
     * @param caps specifys whether device capabilities should be retrieved or not
     * @return true if the device was found, false otherwise
     * @throws DeviceInventoryException
     */
    public boolean getDevice(Sync4jDevice device, boolean caps) throws DeviceInventoryException;

    /**
     * Updates the given device. The capabilities of the device won't chance.
     *
     * @param device Device to be updated
     * @throws DeviceInventoryException
     */
    public void setDevice(Sync4jDevice device) throws DeviceInventoryException;

    /**
     * Update the identifier capabilities of the given device with the
     * specify identifier.
     *
     * @param deviceId the identifier of the device
     * @param capsId   the identifier of the capabilities
     *
     * @throws DeviceInventoryException
     */
    public void setDeviceIdCaps(String deviceId, Long capsId)
    throws DeviceInventoryException;

    /**
     * Retrieves the devices that match the given clause
     *
     * @param clause The additional where clause
     * @return Sync4jDevice[] devices found in DB
     * @throws DeviceInventoryException
     */
    public Sync4jDevice[] queryDevices(Clause clause) throws DeviceInventoryException;

    /**
     * Retrieves the devices that match the given clause specifying whether device capabilities should be retrieved as well.
     *
     * @param clause The additional where clause
     * @param caps   Specifys whether device capabilities should be retrieved
     * @return Sync4jDevice[] devices found in DB
     * @throws DeviceInventoryException
     */
    public Sync4jDevice[] queryDevices(Clause clause, boolean caps) throws DeviceInventoryException;

    /**
     * Returns the number of the devices in the repository that match the given clause filter.
     * If clause is null or an instance of AllClause, count all records
     *
     * @param clause The additional where clause
     * @return int number of devices
     * @throws DeviceInventoryException
     */
    public int countDevices(Clause clause) throws DeviceInventoryException;

    /**
     * Removes the given device and his capabilities
     * @param device the device to remove
     * @throws DeviceInventoryException
     */
    public void removeDevice(Sync4jDevice device) throws DeviceInventoryException;

    /**
     * Removes the given capabilities
     * @param capabilities the capabilities to remove
     * @throws DeviceInventoryException
     */
    public void removeCapabilities(Capabilities capabilities) throws DeviceInventoryException;
}
