/**
 * Copyright (C) 2003-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */

package com.funambol.framework.server.store;

import java.util.Map;

import java.sql.*;

import javax.naming.NamingException;
import javax.sql.DataSource;

import com.funambol.framework.logging.FunambolLogger;
import com.funambol.framework.logging.FunambolLoggerFactory;
import com.funambol.framework.server.store.PersistentStoreException;
import com.funambol.framework.tools.DataSourceTools;
import com.funambol.framework.tools.DBTools;


/**
 * This is a base class for <i>PersistenStore</i> objects. It does not persist
 * any object, but it provides services common to concrete implementations.
 *
 * @version $Id: BasePersistentStore.java,v 1.7 2007/01/18 10:19:10 nichele Exp $
 *
 */
public abstract class BasePersistentStore {

    // --------------------------------------------------------------- Constants

    public static final String CONFIG_JNDI_DATA_SOURCE_NAME = "jndi-data-source-name";

    /**
     * Logger
     */
    protected transient final FunambolLogger log = FunambolLoggerFactory.getLogger();

    // -------------------------------------------------------------- Properties

    /**
     * The database product name
     */
    private String dbProductName = null;

    /**
     * The JNDI name of the datasource to be used
     */
    protected String jndiDataSourceName = null;

    public String getJndiDataSourceName() {
        return this.jndiDataSourceName;
    }

    public void setJndiDataSourceName(String jndiDataSourceName) throws PersistentStoreException {
        this.jndiDataSourceName = jndiDataSourceName;

        if (jndiDataSourceName == null) {
            dataSource = null;
        }

        try {
            dataSource = DataSourceTools.lookupDataSource(jndiDataSourceName);
        } catch (NamingException e) {
            throw new PersistentStoreException("Data source "
            + jndiDataSourceName
            + " not found"
            , e
            );
        }

        Connection conn = null;
        try {
            conn = dataSource.getConnection();
            dbProductName = conn.getMetaData().getDatabaseProductName();
            sqlHelperClause = new SQLHelperClause(dbProductName);
        } catch(SQLException e) {
            log.error("Error getting connection", e);
            throw new PersistentStoreException("Error getting connection (" +
                                               e.getMessage() + ")", e);
        } finally {
            DBTools.close(conn, null, null);
        }
    }

    // ---------------------------------------------------------- Protected data

    protected transient DataSource dataSource = null;
    protected SQLHelperClause sqlHelperClause = null;

    // ------------------------------------------------------------ Constructors

    // ---------------------------------------------------------- Public methods

    /** Configures the persistent store
     *
     * @param config an <i>Map</i> containing configuration parameters.
     *
     * @throws ConfigPersistentStoreException
     *
     */
    public void configure(Map config) throws ConfigPersistentStoreException {

        checkConfigParams(config);

        try {
            setJndiDataSourceName((String) config.get(CONFIG_JNDI_DATA_SOURCE_NAME));
        } catch (PersistentStoreException e) {
            throw new ConfigPersistentStoreException( "Error creating the datasource: "
                                                    + e.getMessage()
                                                    , e
                                                    );
        }
    }

    // ------------------------------------------------------- Protected methods

    // ------------------------------------------------------- Protected methods

    // --------------------------------------------------------- Private methods

    /**
     * Checks if the given configuration parameters contain all required
     * parameters. If not a <i>ConfigPersistentStoreException</i> is thrown.
     *
     * @param config the <i>Map</i> containing the configuration parameters
     *
     * @throws ConfigPersistentStoreException in case of missing parameters
     */
    private void checkConfigParams(Map config)
    throws ConfigPersistentStoreException {
        //
        // Nothing to check
        // 
    }
}

