/**
 * Copyright (C) 2006-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */

package com.funambol.framework.tools.beans;

import java.io.File;
import java.io.IOException;

import java.net.MalformedURLException;
import java.net.URL;

import java.util.HashMap;
import java.util.Map;

import com.funambol.framework.tools.IOTools;
import com.funambol.framework.tools.beans.BeanClassLoader;
import com.funambol.framework.tools.beans.BeanFactory;

/**
 * This class implements a cache of bean objects. Objects are stored in
 * the cache associated with their last modification timestamp. A new instance
 * of the object will be returned only if its server bean is newer  of the one
 * loaded.
 * The bean files are read from <code>configPath<code>.
 *
 * It also provides cached access to singleton instances. In case of singleton
 * istances, a new instance of a server bean is returned only if its
 * configuration has changed; otherwise the last created instance is returned.
 *
 *
 * NOTE: the methods getBeanInstance() and getNewBeanInstance() are not
 *       synchronized by design. We are introducing this class to improve
 *       performance and synchronized methods on frequently used classes
 *       reduce concurrency significantly. From a functionality point of view,
 *       a race condition won't be very negative: only on high load, some
 *       threads for a very little time would still see the old server bean
 *       configuration even if the configuration file has already changed.
 *       This is acceptable considering the gain in performance.
 *
 *
 * @version $Id: BeanCache.java,v 1.2 2007/01/11 11:48:59 nichele Exp $
 */
public class BeanCache {
    // ------------------------------------------------------------ Private data

    /**
     * The hash map containing the cached server beans. Map keys will be the
     * server bean names; the entry will be the corresponding ServerBeanEntry
     * objects.
     */
    private Map cache;

    /**
     * This is the directory where the beans will be read
     */
    private String configPath;

    /**
     * The URL that represents the config path
     */
    private URL configPathURL;

    /**
     * The classLoader used to load the bean
     */
    private ClassLoader classLoader;


    // ------------------------------------------------------------ Constructors

    /**
     * Creates a new BeanCache with the given configuration path
     * @param configPath the config path
     */
    public BeanCache(String configPath) {
        this.cache         = new HashMap();
        this.configPath    = configPath;
        try {
            this.configPathURL = fixURL(configPath);
        } catch (MalformedURLException ex) {
            throw new IllegalArgumentException(
                "Unable to create an URL object using the given configPath"
            );
        }
        this.classLoader   = new BeanClassLoader(configPathURL,
                                                 this.getClass().getClassLoader());
    }

    // ---------------------------------------------------------- Public methods

    /**
     * Returns a new instance of the server bean identified by the given name.
     * @return the server bean instance (a new instance is created each time)
     * @param name the server bean name
     * @throws BeanException in case of errors
     */
    public Object getNewBeanInstance(String name)
    throws BeanException {
        ServerBeanEntry entry = (ServerBeanEntry)cache.get(name);

        try {
            //
            // We first check if the bean is in the configuration path with
            // the given name. If so, the file is the configuration to use,
            // but we need to read it only if recently modified.
            // If the file is not found, the bean represents a class or
            // something else in the classpath that we do not have to deal with.
            //
            String config = null;
            File configFile = new File(configPath, name);

            if (!configFile.exists()) {
                return BeanFactory.getBeanInstance(classLoader, name);
            }

            if ((entry == null) || (entry.ts != configFile.lastModified())) {
                entry = new ServerBeanEntry(configFile);
                cache.put(name, entry);
            }

            return BeanFactory.getBeanInstanceFromConfig(classLoader, entry.config);
        } catch (Exception e) {
            throw new BeanException("Error in creating an instance of "
                                    + name
                                    , e                                );
        }
    }

    /**
     * Returns an instance of the server bean identified by the given name. Once
     * an instance of the object is created, it is stored in the cache for
     * future access. A new instance of the bean is created only when its config
     * file has changed.
     *
     *
     * @return the server bean instance (a new instance is created each time)
     * @param name the server bean name
     * @throws BeanException in case of errors
     */
    public Object getBeanInstance(String name)
    throws BeanException {
        ServerBeanEntry entry = (ServerBeanEntry)cache.get(name);

        try {
            //
            // We first check if the bean is in the configuration path with
            // the given name. If so, the file is the configuration to use,
            // but we need to read it only if recently modified.
            // If the file is not found, the bean represents a class or
            // something else in the classpath that we do not have to deal with.
            //
            String config = null;
            File configFile = new File(configPath, name);

            if (!configFile.exists()) {
                return BeanFactory.getBeanInstance(classLoader, name);
            }

            if ((entry == null) || (entry.ts != configFile.lastModified())) {
                entry = new ServerBeanEntry(configFile);
                entry.instance =
                    BeanFactory.getBeanInstanceFromConfig(classLoader, entry.config);
                cache.put(name, entry);
            }

            return entry.instance;
        } catch (Exception e) {
            throw new BeanException("Error in creating an instance of "
                                    + name
                                    , e                                );
        }
    }

    /**
     * Invalidates the entire cache
     */
    public void invalidate() {
        cache.clear();
    }

    /**
     * Invalidates the given entry
     *
     * @param name the bean name to invalidate
     */
    public void invalidate(String name) {
        cache.remove(name);
    }

    /**
     * Returns the URL that represents the configPath
     * @return the URL that represents the configPath
     */
    public URL getConfigPathURL() {
        return configPathURL;
    }
    // ------------------------------------------------------- Protected methods
    /**
     * Checks if the given string is an accepted URL (starting with http://,
     * https:// or file://). If yes, a new URL object representing the given
     * url is returned; otherwise, the given string is considered a file name
     * and a new URL is obtained calling File.toURL().
     * @param s the string to check
     * @return the corresponding URL if the string represents a URL or the
     *         fixed URL if the string is a pathname/filename
     * @throws MalformedURLException if an error occurs
     */
    protected static URL fixURL(final String s)
    throws MalformedURLException {

        try {
            return new URL(s);
        } catch (MalformedURLException e) {
            //
            // This is not a URL, let's consider it just a file
            //
        }

        try {
            return new File(new File(s).getCanonicalPath()).toURL();
        } catch (IOException e) {
            throw new MalformedURLException("Unable to convert" + s + " to a URL");
        }
    }

    // --------------------------------------------------------- Private classes

    /**
     * This represents an entry in the cache. It associates a server bean name
     * to:
     * <ul>
     * <li>ts -> the timestamp of when the configuration file was last read</li>
     * <li>config -> the configuration file content
     * <li>instance -> the singleton instance associated to this server bean
     *
     */
    class ServerBeanEntry {

        /**
         * file timestamp
         */
        public long ts;
        public String config;
        public Object instance;

        /**
         * Creates a new ServerBeanEntry from the given file
         *
         * @param file the bean file - NOT NULL
         */
        public ServerBeanEntry(File file) {
            ts = file.lastModified();
            try {
                config = IOTools.readFileString(file);
            } catch (Exception e) {
                //
                // just an easy way to let the caller know, even if it is not
                // very elegant.... :)
                //
                config = e.getMessage();
            }
        }
    }
}
