/**
 * Copyright (C) 2003-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.framework.tools.beans;

import java.beans.XMLEncoder;
import java.beans.XMLDecoder;

import java.io.*;
import com.funambol.framework.tools.IOTools;

/**
 * This is a JavaBeans factory. At the moment it has a very simple
 * implementation: it loads the class from a serialized file at runtime.
 *
 * @version $Id: BeanFactory.java,v 1.6 2007/02/04 16:28:48 nichele Exp $
 */
public class BeanFactory {

    protected BeanFactory() { }

    /**
     * Creates an instance of the given bean. The bean is first search as a
     * class. If not found as class, it is searched as a resource. There isn't a
     * lazy initialization.
     *
     * @param       classLoader the class loader to use for loading the bean
     * @param       beanName    name of the bean (for now is also the file name,
     *                          without .ser, into which the bean has been serialized)
     *                          NOT NULL
     *
     * @return      a new instance of the serialized bean
     *
     * @throws      BeanInstantiationException if the bean cannot be istantiated
     * @throws      BeanInitializationException if an initialization error occurred
     * @throws      BeanNotFoundException if the bean does not exist
     */
    public static Object getNoInitBeanInstance(ClassLoader classLoader, String beanName)
    throws BeanInstantiationException,
           BeanInitializationException,
           BeanNotFoundException {

        if (classLoader == null) {
            classLoader = ClassLoader.getSystemClassLoader();
        }

        Object bean = null;

        //
        // Search beanName as a class
        //
        try {
            bean = Class.forName(beanName, true, classLoader).newInstance();
            if (bean instanceof LazyInitBean) {
                ((LazyInitBean)bean).init();
            }
            return bean;
        }  catch (ClassNotFoundException e) {
            //
            // beanName does not represent a class!
            //
        } catch (BeanInitializationException e) {
            throw e;
        } catch (Exception e) {
            //
            // Other exceptions are not good!
            //
            throw new BeanInstantiationException( "Error instantiating "
                                                + beanName
                                                , e
                                                );
        }
        InputStream is = classLoader.getResourceAsStream(beanName);

        if (is == null) {
            throw new BeanNotFoundException( beanName, classLoader);
        }

        BeanExceptionListener exceptionListener = new BeanExceptionListener();

        //
        // IMPORTANT NOTE
        // --------------
        //
        // This serialization is a very important flaw for scalability, but we
        // must do it for a bug in XMLDecoder (see bug #  4822050:
        // http://developer.java.sun.com/developer/bugParade/bugs/4822050.html )
        // 2005-11-23 : it looks like the bug is gone in the 1.5, but it is too
        // early to mandate a JVM 1.5
        //
        synchronized (BeanFactory.class) {
            XMLDecoder d = new XMLDecoder(is, null, exceptionListener);
            bean = d.readObject();
            d.close();
        }

        if (exceptionListener.exceptionThrown()) {
            Throwable t = exceptionListener.getException();

            if (t.getCause() != null) {
                t = t.getCause();
            }
            throw new BeanInstantiationException( "Error instantiating "
                                                + beanName
                                                , t
                                                );
        }

        return bean;
    }

    /**
     * Creates an instance of the given bean. The bean is first search as a
     * class. If not found as class, it is searched as a resource.
     *
     * @param       classLoader the class loader to use for loading the bean
     * @param       beanName    name of the bean (for now is also the file name,
     *                          without .ser, into which the bean has been serialized)
     *                          NOT NULL
     *
     * @return      a new instance of the serialized bean
     *
     * @throws      BeanInstantiationException if the bean cannot be istantiated
     * @throws      BeanInitializationException if an initialization error occurred
     * @throws      BeanNotFoundException if the bean does not exist
     */
    public static Object getBeanInstance(ClassLoader classLoader, String beanName)
    throws BeanInstantiationException,
           BeanInitializationException,
           BeanNotFoundException {

        Object bean = null;

        bean = getNoInitBeanInstance(classLoader, beanName);

        if (bean instanceof LazyInitBean) {
            ((LazyInitBean)bean).init();
        }

        return bean;
    }


    /**
     * Creates an instance of the given bean using the system class loader.
     *
     * @param       beanName    name of the bean (for now is also the file name,
     *                          without .ser, into which the bean has been serialized)
     *
     * @return      a new instance of the serialized bean
     *
     * @throws      BeanInstantiationException if the bean cannot be istantiated
     * @throws      BeanInitializationException if an initialization error occurred
     * @throws      BeanNotFoundException if the bean does not exist
     */
    public static Object getBeanInstance(String beanName)
    throws BeanInstantiationException,
           BeanInitializationException,
           BeanNotFoundException {
           return getBeanInstance(null, beanName);
    }


    /**
     * Creates an instance of the given bean. This version deserializes the bean
     * from the given file.
     *
     * @param       beanFile    the filename of the serialized file
     *
     * @return      a new instance of the serialized bean
     *
     * @throws      BeanInstantiationException if the bean cannot be istantiated
     * @throws      BeanInitializationException if an initialization error occurred
     * @throws      BeanNotFoundException if the bean does not exist
     */
    public static Object getBeanInstance(File beanFile)
    throws BeanInstantiationException,
           BeanInitializationException,
           BeanNotFoundException {
        try {
            Object bean = null;

            //
            // IMPORTANT NOTE
            // --------------
            //
            // This serialization is a very important flaw for scalability, but we
            // must do it for a bug in XMLDecoder (see bug #  4822050:
            // http://developer.java.sun.com/developer/bugParade/bugs/4822050.html )
            // 2005-11-23 : it looks like the bug is gone in the 1.5, but it is too
            // early to mandate a JVM 1.5
            //
            synchronized (BeanFactory.class) {
                XMLDecoder e = new XMLDecoder(new FileInputStream(beanFile));
                bean = e.readObject();
                e.close();
            }

            if (bean instanceof LazyInitBean) {
                ((LazyInitBean)bean).init();
            }
            return bean;
        } catch (FileNotFoundException e) {
            throw new BeanNotFoundException(beanFile.getName(), (ClassLoader)null);
        } catch (Exception e) {
            String msg = "Bean creation ("
                       +          beanFile
                       +       ") failed: "
                       +     e.toString()
                       ;
            throw new BeanInstantiationException(msg, e);
        }
    }

    /**
     * Creates an instance of the given bean. <code>config</code> is the server
     * bean  configuration.
     *
     * @param       classLoader the class loader to use for loading the bean
     * @param       beanConfig  configuration of the server bean to create - NOT NULL
     *
     * @return      a new instance of the serialized bean
     *
     * @throws      BeanInstantiationException if the bean cannot be istantiated
     * @throws      BeanInitializationException if an initialization error occurred
     * @throws      BeanNotFoundException if the bean does not exist
     */
    public static Object getBeanInstanceFromConfig(ClassLoader classLoader, String beanConfig)
    throws BeanInstantiationException,
           BeanInitializationException,
           BeanException {

        Object bean = null;

        bean = unmarshal(classLoader, beanConfig);

        if (bean instanceof LazyInitBean) {
            ((LazyInitBean)bean).init();
        }

        return bean;
    }

    /**
     * Creates an instance of the given bean. <code>config</code> is the server
     * bean  configuration. It is a shortcut for getBeanInstanceFromConfig(null, config)
     *
     * @param       beanConfig  configuration of the server bean to create - NOT NULL
     *
     * @return      a new instance of the serialized bean
     *
     * @throws      BeanInstantiationException if the bean cannot be istantiated
     * @throws      BeanInitializationException if an initialization error occurred
     * @throws      BeanNotFoundException if the bean does not exist
     */
    public static Object getBeanInstanceFromConfig(String beanConfig)
    throws BeanInstantiationException,
           BeanInitializationException,
           BeanException {

        Object bean = null;

        bean = unmarshal(null, beanConfig);

        if (bean instanceof LazyInitBean) {
            ((LazyInitBean)bean).init();
        }

        return bean;
    }

    /**
     * Returns the server bean configuration if there is any. If the bean name
     * does not represents a configuration but an instance, null is returned.
     *
     * @param       classLoader the class loader to use for loading the bean
     * @param       beanName    name of the bean - NOT NULL
     *
     * @return      the configuration file if found, NULL otherwise
     *
     * @throws      BeanInstantiationException if the bean cannot be istantiated
     * @throws      BeanInitializationException if an initialization error occurred
     * @throws      BeanNotFoundException if the bean does not exist
     */
    public static String getBeanConfig(ClassLoader classLoader, String beanName)
    throws BeanInstantiationException,
           BeanInitializationException,
           BeanNotFoundException {
        if (classLoader == null) {
            classLoader = ClassLoader.getSystemClassLoader();
        }

        InputStream is = classLoader.getResourceAsStream(beanName);

        if (is == null) {
            return null;
        }

        try {
            return IOTools.readFileString(is);
        } catch (IOException e) {
            throw new BeanNotFoundException("The bean configuration could not be read: " + e.toString());
        } finally {
            try {
                is.close();
            } catch (Exception e) {}
        }
    }

    /**
     * Serialize the given object in the give file using XMLEncoder.
     *
     * @param obj the object to be serialized
     * @param file the file into wich serialize the object
     *
     * @throws BeanException in case of errors
     */
    public static void saveBeanInstance(Object obj, File file)
    throws BeanException {

        File parent = file.getParentFile();
        if (!parent.exists()) {
            parent.mkdirs();
        }

        XMLEncoder encoder = null;

        try {
            encoder = new XMLEncoder(new FileOutputStream(file));
            encoder.writeObject(obj);
        } catch (IOException e) {
            String msg = "Bean saving (" + file + ") failed: " + e.toString();
            throw new BeanException(msg, e);
        } finally {
            if (encoder != null) {
                encoder.close();
            }
        }
    }

    /**
     * Marshals the given object into XML using <i>java.beans.XMLEncoder</i>.
     *
     * @param o the object to marshal
     *
     * @return the XML form of the object
     *
     * @throws BeanException in case of marshalling errors
     */
    public static String marshal(ClassLoader classLoader, Object o)
    throws BeanException {
        XMLEncoder enc = null;
        ClassLoader oldClassLoader = null;

        if (classLoader != null) {
            oldClassLoader = Thread.currentThread().getContextClassLoader();
            Thread.currentThread().setContextClassLoader(classLoader);
        }

        BeanExceptionListener exceptionListener = new BeanExceptionListener();

        try {
            ByteArrayOutputStream os = new ByteArrayOutputStream();

            enc = new XMLEncoder(os);

            enc.writeObject(o);

            enc.close(); enc = null; // we need to do this here to get the string

            if (exceptionListener.exceptionThrown()) {
                Throwable t = exceptionListener.getException();
                throw new BeanException(t.toString(), t);
            }

            return os.toString();
        } finally {
            if (enc != null) {
                enc.close();
            }

            if (classLoader != null) {
                Thread.currentThread().setContextClassLoader(oldClassLoader);
            }
        }
    }

    /**
     * Marshals the given object into XML using <i>java.beans.XMLEncoder</i>.
     *
     * @param o the object to marshal
     *
     * @return the XML form of the object
     *
     * @throws BeanException in case of marshalling errors
     */
    public static String marshal(Object o)
    throws BeanException {
        return marshal(null, o);
    }


    /**
     * Unmarshals the given serialized object into the corresponding instance
     * using <i>java.beans.XMLDecoder</i>.
     *
     * @param classLoader the class loader to use (null for the standard one)
     * @param xml XML serialized form
     *
     * @return the unmarshalled instance
     *
     * @throws BeanException in case of unmarshalling errors
     */
    public static Object unmarshal(ClassLoader classLoader, String xml)
    throws BeanException {
        XMLDecoder dec = null;
        ClassLoader oldClassLoader = null;

        if (classLoader != null) {
            oldClassLoader = Thread.currentThread().getContextClassLoader();
            Thread.currentThread().setContextClassLoader(classLoader);
        }

        BeanExceptionListener exceptionListener = new BeanExceptionListener();

        try {
            ByteArrayInputStream is = new ByteArrayInputStream(xml.getBytes());
            Object ret = null;

            dec = new XMLDecoder(is);
            dec.setExceptionListener(exceptionListener);
            ret = dec.readObject();
            dec.close(); dec = null;

            if (exceptionListener.exceptionThrown()) {
                Throwable t = exceptionListener.getException();
                throw new BeanException(t.toString(), t);
            }

            return ret;
        } finally {
            if (dec != null) {
                dec.close();
            }

            if (classLoader != null) {
                Thread.currentThread().setContextClassLoader(oldClassLoader);
            }
        }
    }

    /**
     * The same as <i>unmarshal(null, xml)</i>
     *
     * @param xml XML serialized form
     *
     * @return the unmarshalled instance
     *
     * @throws BeanException in case of unmarshalling errors
     */
    public static Object unmarshal(String xml)
    throws BeanException {
        return unmarshal(null, xml);
    }

    /**
     * Creates a bean instance given the class name and serialize it in the
     * specified file. The object is serialized in XML.
     * <p>
     * Syntax:<br>
     * com.funambol.framework.tools.beans.BeanFactory <i>class name</i> </i>file name</i>
     **/
    public static void main(String[] args) throws Exception {
        if (args.length != 2) {
            System.out.println("Syntax: com.funambol.framework.tools.beans.BeanFactory <class name> <file name>");
            return;
        }

        saveBeanInstance(getBeanInstance(args[0]), new File(args[1]));
    }
}
