/**
 * Copyright (C) 2006-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */

package com.funambol.framework.tools.beans;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

/**
 * Useful class to handle bean files. It's handled as singleton class with a factory
 * method to obtain a <CODE>BeanTool</CODE> instance to handle a given configPath.
 * Any <CODE>BeanTool</CODE> instance uses a <CODE>BeanCache</CODE> instance to
 * cache the rad bean files.
 * @version $Id: BeanTool.java,v 1.2 2007/01/11 11:48:59 nichele Exp $
 */
public class BeanTool {

    /**
     * the BeanCache instance to cache the bean files
     */
    private BeanCache beanCache = null;

    /**
     * the BeanTool instances
     */
    private static Map<URL, BeanTool> instances = new HashMap<URL, BeanTool>();

    /**
     * Creates a new instance of BeanTool to handle bean files in the given path
     * @param path the path where to read bean files
     */
    private BeanTool(String path) {
        beanCache = new BeanCache(path);
    }

    // ---------------------------------------------------------- Public methods
    /**
     * Returns a <CODE>BeanTool</CODE> instance to handle bean file in the given path
     * @param path the path where to read the bean files
     * @return a BeanTool instance
     */
    public static synchronized BeanTool getBeanTool(String path) {
        URL url = null;
        try {
            url = BeanCache.fixURL(path);
        } catch (MalformedURLException ex) {
            throw new IllegalArgumentException(
                "Unable to create an URL object using the given configPath"
            );
        }
        BeanTool beanTool = instances.get(url);
        if (beanTool == null) {
            beanTool = new BeanTool(path);
            instances.put(url, beanTool);
        }
        return beanTool;
    }

    /**
     * Returns a new instance of the server bean identified by the given name.
     * @return the server bean instance (a new instance is created each time)
     * @param name the server bean name
     * @throws com.funambol.framework.tools.beans.BeanException if an error occurs
     */
    public Object getNewBeanInstance(String name)
    throws BeanException {
        return beanCache.getNewBeanInstance(name);
    }

    /**
     * Returns an instance of the server bean identified by the given name. Once
     * an instance of the object is created, it is stored in the cache for
     * future access. A new instance of the bean is created only when its config
     * file has changed.
     * @return the server bean instance (a new instance is created each time)
     * @param name the server bean name
     * @throws com.funambol.framework.tools.beans.BeanException if an error occurs
     */
    public Object getBeanInstance(String name)
    throws BeanException {
        return beanCache.getBeanInstance(name);
    }

}
