/**
 * Copyright (C) 2003-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.server.admin.ejb;

import java.util.Map;

import javax.ejb.CreateException;
import javax.ejb.EJBException;
import javax.ejb.SessionBean;
import javax.ejb.SessionContext;

import org.apache.log4j.Appender;

import com.funambol.framework.config.LoggerConfiguration;

import com.funambol.framework.config.LoggingConfiguration;
import com.funambol.framework.core.Alert;
import com.funambol.framework.engine.source.SyncSource;
import com.funambol.framework.filter.Clause;
import com.funambol.framework.security.Sync4jPrincipal;
import com.funambol.framework.server.Capabilities;
import com.funambol.framework.server.LastTimestamp;
import com.funambol.framework.server.Sync4jDevice;
import com.funambol.framework.server.Sync4jModule;
import com.funambol.framework.server.Sync4jSource;
import com.funambol.framework.server.Sync4jUser;
import com.funambol.framework.server.error.ServerException;

import com.funambol.server.admin.AdminException;
import com.funambol.server.admin.AdminManager;
import com.funambol.server.config.Configuration;
import com.funambol.server.config.ServerConfiguration;

/**
 * This is the session enterprise java bean that handles the administration
 * console. It is designed to be a stateless session bean.
 * <p>
 *  This server accepts the requests addressed to the hostname
 *  indicated by the configuration property pointed by {CONFIG_SERVER_URI} (see
 *  Funambol.xml).
 * <p>
 * AdminBean uses the system property funambol.configpath to set the base of the
 * config path.
 *
 * @version $Id: AdminBean.java,v 1.10 2007/02/25 18:49:34 nichele Exp $
 *
 */
public class AdminBean implements SessionBean {

    // ------------------------------------------------------------ Private data
    private SessionContext sessionContext = null;
    private AdminManager   admin          = null;

    // ------------------------------------------------------------- EJB methods
    public void ejbCreate() throws CreateException {
        admin = new AdminManager();
    }

    public Configuration getConfig() {
        return admin.getConfig();
    }

    public void ejbRemove() {
    }

    public void ejbActivate() {
    }

    public void ejbPassivate() {
    }

    public void setSessionContext(SessionContext sessionContext)
    throws EJBException {
        this.sessionContext = sessionContext;
    }

    /**
     * Read the list of roles available.
     *
     * @return names of roles available
     *
     * @throws ServerException
     * @throws AdminException
     */
    public String[] getRoles()
    throws ServerException, AdminException {
        return admin.getRoles();
    }

    /**
     * Read all users that satisfy the parameter of search.
     *
     * @param clause array of conditions for the query
     *
     * @return array of Sync4jUser
     *
     * @throws ServerException
     * @throws AdminException
     */
    public Sync4jUser[] getUsers(Clause clause)
    throws ServerException, AdminException {
        return admin.getUsers(clause);
    }

    /**
     * Adds a new user and the assigned role to it
     *
     * @param user the user to add
     *
     * @throws ServerException
     * @throws AdminException
     */
    public void addUser(Sync4jUser user)
    throws ServerException, AdminException {
        admin.addUser(user);
    }

    /**
     * Update the information of the specific user
     *
     * @param user the user with informations updated
     *
     * @throws ServerException
     * @throws AdminException
     */
    public void setUser(Sync4jUser user)
    throws ServerException, AdminException {
        admin.setUser(user);
    }

    /**
     * Delete the user
     *
     * @param userName the name of user to delete
     *
     * @throws ServerException
     * @throws AdminException
     */
    public void deleteUser(String userName)
    throws ServerException, AdminException {
        admin.deleteUser(userName);
    }

    /**
     * Count the number of users that satisfy the specif clauses.
     *
     * @param clause the conditions of the search
     *
     * @return number of users
     *
     * @throws ServerException
     * @throws AdminException
     */
    public int countUsers(Clause clause)
    throws ServerException, AdminException {
        return admin.countUsers(clause);
    }

    /**
     * Read a list of device that satisfy the specific conditions.
     *
     * @param clauses the conditions foe search informations.
     *
     * @return array of device
     *
     * @throws ServerException
     * @throws AdminException
     */
    public Sync4jDevice[] getDevices(Clause clauses)
    throws ServerException, AdminException {
        return admin.getDevices(clauses);
    }

    /**
     * Adds a new device.
     *
     * @param d the new device
     *
     * @return the device id
     *
     * @throws ServerException
     * @throws AdminException
     */
    public String addDevice(Sync4jDevice d)
    throws ServerException, AdminException {
        return admin.addDevice(d);
    }

    /**
     * Retrieves the device with the given id.
     *
     * @param id the device id
     *
     * @return the retrieved device or null if no devices were found
     *
     * @throws ServerException
     * @throws AdminException
     */
    public Sync4jDevice getDevice(final String id)
    throws ServerException, AdminException {
        return admin.getDevice(id);
    }

    /**
     * Read the device's capabilities for the specific device.
     *
     * @param deviceId the identifier of device
     *
     * @return the device's capabilities
     *
     * @throws ServerException
     * @throws AdminException
     */
    public Capabilities getDeviceCapabilities(String deviceId)
    throws ServerException, AdminException {
        return admin.getDeviceCapabilities(deviceId);
    }

    /**
     * Save the device's capabilities for the specific device.
     *
     * @param caps     the Capabilities to insert
     * @param deviceID the identifier of device
     *
     * @return the identifier of the inserted capabilities
     *
     * @throws ServerException
     * @throws AdminException
     */
    public Long setDeviceCapabilities(Capabilities caps, String deviceId)
    throws ServerException, AdminException {
        admin.setDeviceCapabilities(caps, deviceId);
        return caps.getId();
    }


    /**
     * Update the information of specific device.
     *
     * @param d the device to update
     *
     * @throws ServerException
     * @throws AdminException
     */
    public void setDevice(Sync4jDevice d)
    throws ServerException, AdminException {
        admin.setDevice(d);
    }

    /**
     * Delete the specific device.
     *
     * @param deviceId the device id that identifies the device
     *
     * @throws ServerException
     * @throws AdminException
     */
    public void deleteDevice(String deviceId)
    throws ServerException, AdminException{
        admin.deleteDevice(deviceId);
    }

    /**
     * Count the number of device that satisfy the specific clauses.
     *
     * @param clauses the specific conditions for search device
     *
     * @return the number of device finds
     *
     * @throws ServerException
     * @throws AdminException
     */
    public int countDevices(Clause clauses)
    throws ServerException, AdminException {
        return admin.countDevices(clauses);
    }

    /**
     * Read all principal that satisfy the clauses.
     *
     * @param clauses the specific conditions for search principal
     *
     * @return array of principal
     *
     * @throws ServerException
     * @throws AdminException
     */
    public Sync4jPrincipal[] getPrincipals(Clause clauses)
    throws ServerException, AdminException {
        return admin.getPrincipals(clauses);
    }

    /**
     * Adds a new principal.
     *
     * @param p the new principal
     *
     * @return the principal id
     *
     * @throws ServerException
     * @throws AdminException
     */
    public long addPrincipal(Sync4jPrincipal p)
    throws ServerException, AdminException {
        return admin.addPrincipal(p);
    }

    /**
     * Delete the specific principal.
     *
     * @param principalId the principal id that identifies the principal
     *
     * @throws ServerException
     * @throws AdminException
     */
    public void deletePrincipal(long principalId)
    throws ServerException, AdminException {
        admin.deletePrincipal(principalId);
    }

    /**
     * Count the number of principal that satisfy the specific clauses.
     *
     * @param clauses the specific conditions for search principal
     *
     * @return the number of principal finds
     *
     * @throws ServerException
     * @throws AdminException
     */
    public int countPrincipals(Clause clauses)
    throws ServerException, AdminException {
        return admin.countPrincipals(clauses);
    }

    /**
     * Read all last timestamps that satisfy the clauses.
     *
     * @param clauses the specific conditions for search last timestamps
     *
     * @return array of last timestamps
     *
     * @throws ServerException
     * @throws AdminException
     */
    public LastTimestamp[] getLastTimestamps(Clause clauses)
    throws ServerException, AdminException {
        return admin.getLastTimestamps(clauses);
    }

    /**
     * Delete the specific last timestamps.
     *
     * @param principalId the principal id that identifies the  principal
     * @param sourceId the source id that identifies the  source
     * @throws ServerException
     * @throws AdminException
     */
    public void deleteLastTimestamp(long principalId, String sourceId)
    throws ServerException, AdminException {
        admin.deleteLastTimestamp(principalId, sourceId);
    }

    /**
     * Count the number of last timestamp that satisfy the specific clauses.
     *
     * @param clauses the specific conditions for search last timestamp
     *
     * @return the number of last timestamp finds
     *
     * @throws ServerException
     * @throws AdminException
     */
    public int countLastTimestamps(Clause clauses)
    throws ServerException, AdminException {
        return admin.countLastTimestamps(clauses);
    }

    /**
     * Read all modules information
     *
     * @return an array with the modules information (empty if no objects are found)
     *
     * @throws ServerException
     * @throws AdminException
     */
    public Sync4jModule[] getModulesName()
    throws ServerException, AdminException {
        return admin.getModulesName();
    }

    /**
     * Read the information of the specific module
     *
     * @param moduleId the module id that identifies the module to search
     *
     * @return the relative information to the module
     *
     * @throws ServerException
     * @throws AdminException
     */
    public Sync4jModule getModule(String moduleId)
    throws ServerException, AdminException {
        return admin.getModule(moduleId);
    }

    /**
     * Adds a new source into datastore and create a relative file xml with configuration.
     * The source must have a defined source type.
     * The source type must refer to a connector.
     * The connector must refer to a module.
     *
     * @param moduleId the module id
     * @param connectorId the connector id
     * @param sourceTypeId the source type id
     * @param source the information of the new source
     *
     * @throws ServerException
     * @throws AdminException
     */
    public void addSource(String moduleId, String connectorId, String sourceTypeId, SyncSource source)
    throws ServerException, AdminException {
        admin.addSource(moduleId, connectorId, sourceTypeId, source);
    }

    /**
     * Read a list of syncSource that satisfy the specific conditions.
     *
     * @param clauses the conditions foe search informations.
     *
     * @return array of syncSource
     *
     * @throws ServerException
     * @throws AdminException
     */
    public Sync4jSource[] getSync4jSources(Clause clauses)
    throws ServerException, AdminException {
        return admin.getSync4jSources(clauses);
    }

    /**
     * Returns an array with the classes relative to the given source types
     *
     * @param sourceTypesId of source types identifier
     *
     * @return array of class names
     *
     * @throws ServerException
     * @throws AdminException
     */
    public String[] getSyncSourceClasses(String[] sourceTypesId)
    throws ServerException, AdminException {
        return admin.getSyncSourceClasses(sourceTypesId);
    }

    /**
     * Update a specific source into datastore and create a relative file xml with configuration.
     * The source must have a defined source type.
     * The source type must refer to a connector.
     * The connector must refer to a module.
     *
     * @param moduleId the module id
     * @param connectorId the connector id
     * @param sourceTypeId the source type id
     * @param source the information of the new source
     *
     * @throws ServerException
     * @throws AdminException
     */
    public void setSource(String moduleId, String connectorId, String sourceTypeId, SyncSource source)
    throws ServerException, AdminException{
        admin.setSource(moduleId, connectorId, sourceTypeId, source);
    }

    /**
     * Delete a specific source and the relative file of configuration.
     *
     * @param sourceUri the uri that identifies the source
     *
     * @throws ServerException
     * @throws AdminException
     */
    public void deleteSource(String sourceUri)
    throws ServerException, AdminException{
        admin.deleteSource(sourceUri);
    }

    /**
     * Returns an array with all available <code>LoggerConfiguration</code> objects
     * on the server.
     *
     * @return an array with all available <code>LoggerConfiguration</code>
     *         objects on the server.
     *
     * @throws ServerException, AdminException
     */
    public LoggerConfiguration[] getLoggers()
    throws ServerException, AdminException {
        return admin.getLoggers();
    }

    /**
     * Returns a map with all available <code>Appender</code> objects
     * on the server.
     *
     * @return a map with all available <code>Appender</code>
     *         objects on the server.
     *
     * @throws ServerException, AdminException
     */
    public Map getAppenders()
    throws ServerException, AdminException {
        return admin.getAppenders();
    }

    /**
     * Sets the given appender
     * @param appender the appender to set
     */
    public void setAppender(Appender appender)
    throws ServerException, AdminException {
        admin.setAppender(appender);
    }

    /**
     * Returns the class name of the management panel to configure an appender
     * with the given class name.
     *
     * @param the appender class name
     * @return the class name of the management panel
     */
    public String getAppenderManagementPanel(String appenderClassName)
    throws ServerException, AdminException {
        return admin.getAppenderManagementPanel(appenderClassName);
    }

    /**
     * Saves and apply the new logging configuration
     *
     * @param settings the new logging configuration
     *
     * @throws ServerException
     * @throws AdminException
     */
    public void setLoggingConfiguration(LoggingConfiguration settings)
    throws ServerException, AdminException {
        admin.setLoggingConfiguration(settings);
    }

    /**
     * Saves and apply the new logger configuration
     *
     * @param loggerConfig the new logger configuration
     *
     * @throws ServerException
     * @throws AdminException
     */
    public void setLoggerConfiguration(LoggerConfiguration loggerConfig)
    throws ServerException, AdminException {
        admin.setLoggerConfiguration(loggerConfig);
    }

    /**
     * Returns the server configuration
     *
     * @return the server configuration as a <i>ServerConfiguration</i>
     *         object.
     *
     * @throws ServerException, AdminException
     */
    public ServerConfiguration getServerConfiguration()
    throws ServerException, AdminException {
        return admin.getServerConfiguration();
    }

    /**
     * Saves and apply the new server configuration
     *
     * @param serverConfig the new server configuration
     *
     * @throws ServerException
     * @throws AdminException
     */
    public void setServerConfiguration(ServerConfiguration serverConfig)
    throws ServerException, AdminException {
        admin.setServerConfiguration(serverConfig);
    }

    /**
     * Returns the server version.
     *
     * @return the server version
     *
     */
    public String getServerVersion()
    throws ServerException, AdminException {
        return admin.getServerVersion();
    }

    /**
     * Returns the server bean with the given name. If the bean does not exist,
     * am AdminException is thrown.
     *
     * @param bean the server bean
     *
     * @throws ServerException, AdminException
     */
    public Object getServerBean(String bean)
    throws ServerException, AdminException {
        return admin.getServerBean(bean);
    }

    /**
     * Sets the server bean with the given name.
     *
     * @param bean the server bean
     * @param obj the bean instance
     *
     * @throws ServerException, AdminException
     */
    public void setServerBean(String bean, Object obj)
    throws ServerException, AdminException {
        admin.setServerBean(bean, obj);
    }

    /**
     * Sends a notification message to the given device.
     *
     * @param deviceId device identifier
     * @param alerts   array of Alert command for the datastore to sync, used
     *                 to create the binary message to send
     * @param uimode   user intraction mode
     */
    public void sendNotificationMessage(String  deviceId,
                                        Alert[] alerts  ,
                                        int     uimode  )
    throws ServerException {

        admin.sendNotificationMessage(deviceId, alerts, uimode);
    }

    /**
     * Sends a notification message to all devices of the principals with the
     * given username.
     *
     * @param username the username
     * @param alerts   array of Alert command for the datastore to sync, used
     *                 to create the binary message to send
     * @param uimode   user intraction mode
     */
    public void sendNotificationMessages(String  username,
                                         Alert[] alerts  ,
                                         int     uimode  )
    throws ServerException {

        admin.sendNotificationMessages(username, alerts, uimode);
    }

}

