/**
 * Copyright (C) 2003-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.server.admin.ws.axis;

import org.apache.axis.AxisFault;
import org.apache.axis.MessageContext;
import org.apache.axis.handlers.BasicHandler;
import org.apache.axis.transport.http.HTTPConstants;
import org.apache.axis.utils.Messages;

import com.funambol.framework.logging.FunambolLogger;
import com.funambol.framework.logging.FunambolLoggerFactory;
import com.funambol.framework.server.Sync4jUser;
import com.funambol.framework.server.store.NotFoundException;
import com.funambol.framework.server.store.PersistentStoreException;

import com.funambol.server.admin.UserManager;

import com.funambol.server.config.Configuration;
import com.funambol.server.config.ConfigurationConstants;

/**
 * This is an Axis authentication handler used to authenticate the user trying
 * to perform administration tasks on the Admin web service. Authentication is
 * currently performed with the Basic scheme.
 *
 *
 * @version $Id: AdminAuthHandler.java,v 1.6 2007/01/11 11:49:00 nichele Exp $
 */
public class AdminAuthHandler
extends BasicHandler
implements ConfigurationConstants {

    // --------------------------------------------------------------- Constants
    public static final String LOG_NAME   = "admin";
    public static final String ROLE_ADMIN = "sync_administrator";

    public static final String FAULT_UNAUTHORIZED = "Server.Unauthenticated";

    private static final String AUTH_ERROR_MESSAGE =
        "Authentication failed, because the server is unable to access to the datastore. Please see server log for the causes.";

    // ------------------------------------------------------------ Private data
    private FunambolLogger log = FunambolLoggerFactory.getLogger(LOG_NAME);
    private UserManager userManager;

    // ------------------------------------------------------------ Constructors
    public AdminAuthHandler() {
        try {
            Configuration c = Configuration.getConfiguration();
            userManager = (UserManager)c.getUserManager();
        } catch (Exception e) {
            log.error("Error retrieving the user manager", e);
        }
    }

    // ---------------------------------------------------------- Public methods

    /**
     * Redefines the BasicHandler's invoke method.
     *
     * @param msgContext the Axis message context
     *
     * @throws AxisFault if authentication fails.
     */
    public void invoke(MessageContext msgContext) throws AxisFault {
        if (log.isTraceEnabled()) {
            log.trace("Authenticating admin action.");
        }

        String username = msgContext.getUsername();
        if (username == null) {
            //
            // Hey! no auth here!
            // This will make axis return a 401
            //
            if (log.isTraceEnabled()) {
                log.trace("Credentials not provided.");
            }
            AxisFault f = new AxisFault();
            f.setFaultCodeAsString("Server.Unauthenticated");
            throw f;
        }

        String password = msgContext.getPassword();
        if (password == null) {
            password = "";
        }

        Sync4jUser user = null;
        try {
            //
            // First of all: is the user authenticated?
            //
            user = new Sync4jUser();
            user.setUsername(username);

            userManager.getUser(user);

            if (!password.equals(user.getPassword())) {
                //
                // Authentication failed!
                //
                notAuthorized();
            }

            //
            // Is the user a sync admin user?
            //
            userManager.getUserRoles(user);

            if (!user.hasRole(ROLE_ADMIN)) {
                //
                // The user is not an administrator
                //
                if (log.isInfoEnabled()) {
                    log.info( "Authorization denied: user "
                            + user
                            + " is not an administrator."
                            );
                }
                notAuthorized();
            }
        } catch (NotFoundException e) {

                if (log.isInfoEnabled()) {
                    log.info( "Authorization denied: user "
                            + user
                            + " not found."
                            );
                }
            notAuthorized();

        } catch (AxisFault f) {
            //
            // this happens, for example, when the user is no more authorized
            //
            throw f;
        } catch (Exception e) {
            log.error("Error retrieving the user", e);

            AxisFault f = new AxisFault();
            f.setFaultString("The server is unable to authenticate the user. Check the server log for details.");
            throw f;
        }

    }

    // --------------------------------------------------------- Private methods

    /**
     * Throws the AxisFault representing "not authorized".
     *
     * @throws AxisFault
     */
    private void notAuthorized()
    throws AxisFault {
        AxisFault f = new AxisFault();
        
        f.setFaultCodeAsString(FAULT_UNAUTHORIZED);
        throw f;
    }
}
