/**
 * Copyright (C) 2003-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.server.config;

import java.io.File;
import java.io.IOException;

import java.net.URL;
import java.util.ArrayList;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Appender;
import org.apache.log4j.FileAppender;
import org.apache.log4j.spi.OptionHandler;

import com.funambol.framework.config.ConfigClassLoader;
import com.funambol.framework.config.ConfigurationException;
import com.funambol.framework.config.LoggerConfiguration;
import com.funambol.framework.config.LoggingConfiguration;

import com.funambol.framework.engine.SyncStrategy;
import com.funambol.framework.engine.pipeline.InputMessageProcessor;
import com.funambol.framework.engine.pipeline.OutputMessageProcessor;
import com.funambol.framework.engine.pipeline.PipelineManager;
import com.funambol.framework.engine.transformer.DataTransformerManager;

import com.funambol.framework.logging.AppenderType;
import com.funambol.framework.logging.AppendersCache;
import com.funambol.framework.logging.FunambolLogger;
import com.funambol.framework.logging.FunambolLoggerFactory;

import com.funambol.framework.security.Officer;

import com.funambol.framework.server.inventory.DeviceInventory;
import com.funambol.framework.server.session.SessionHandler;
import com.funambol.framework.server.store.PersistentStore;

import com.funambol.framework.tools.beans.BeanException;

import com.funambol.server.admin.UserManager;
import com.funambol.server.config.ConfigurationConstants;
import com.funambol.server.store.PersistentStoreManager;
import com.funambol.server.tools.directorymonitor.DirectoryMonitor;
import com.funambol.server.tools.directorymonitor.FileChangeEvent;
import com.funambol.server.tools.directorymonitor.FileChangeEvent.EventType;
import com.funambol.server.tools.directorymonitor.FileChangeListener;

/**
 * Incapsulates all the configuration information about the Sync4j Engine.
 * Configuration parameters are stored as properties in a Map structure. The
 * key is the name of the parameter and it is strucured in dotted separated
 * namespaces. The value of the parameter is the value of the key and can be of
 * any type. Accessor methods are provided to get the value of the parameter in
 * a particular type (such as String, int, double, boolean, ...).<br>
 * Access to getXXX() and setXXX() methods are synchronized.
 * <p>
 * This class follows the Singleton pattern, therefore it cannot be directly
 * instantiated; an internal instance is created just once and returned at any
 * getConfiguration() call.
 *
 * @version $Id: Configuration.java,v 1.26 2007/03/08 14:29:23 nichele Exp $
 */
public class Configuration
implements java.io.Serializable, ConfigurationConstants, FileChangeListener {

    // --------------------------------------------------------------- Constants


    /** The logging directory */
    public static final File DIRECTORY_LOGGING =
        new File(getConfigPath() + File.separator + PATH_LOGGING);

    /** The loggers directory */
    public static final File DIRECTORY_LOGGER =
        new File(getConfigPath() + File.separator + PATH_LOGGER);

    /** The appenders directory */
    public static final File DIRECTORY_APPENDER =
        new File(getConfigPath() + File.separator + PATH_APPENDER);

    /** The persistent stores directory */
    public static final File DIRECTORY_PERSISTENT_STORE =
        new File(getConfigPath() + File.separator + PATH_PERSISTENT_STORE);

    /** The input synclets directory */
    public static final File DIRECTORY_SYNCLET_INPUT =
        new File(getConfigPath() + File.separator + PATH_SYNCLET_INPUT);

   /** The output synclets directory */
    public static final File DIRECTORY_SYNCLET_OUTPUT =
        new File(getConfigPath() + File.separator + PATH_SYNCLET_OUTPUT);


    // ------------------------------------------------------------ Private data

    /**
     * The server configuration
     */
    private static Configuration singleton;

    /** The configured PersistentStoreManager */
    private PersistentStoreManager persistentStoreManager = null;

    /** The configured PipelineManager */
    private PipelineManager pipelineManager = null;

    /** The class loader for the configuration files */
    private transient ClassLoader classLoader = null;

    /** The logger */
    private static transient FunambolLogger log =
        FunambolLoggerFactory.getLogger("configuration");

    /** The monitor of the configuration directory  */
    private DirectoryMonitor directoryMonitor = null;

    // -------------------------------------------------------------- properties

    /**
     * Getter for property classLoader.
     * @return Value of property classLoader.
     */
    public ClassLoader getClassLoader() {

        return classLoader;
    }

    /**
     * Setter for property classLoader.
     * @param classLoader New value of property classLoader.
     */
    public void setClassLoader(ClassLoader classLoader) {
        this.classLoader = classLoader;
    }

    // ------------------------------------------------------------ Constructors

    /**
     * This is disabled for normal use since this class is a singleton.
     */
    protected Configuration() {

        String configPath = getConfigPath();

        directoryMonitor = new DirectoryMonitor(configPath,
                                                DIRECTORY_MONITOR_SCAN_PERIOD);

        directoryMonitor.addFileChangeListener(this);

        DIRECTORY_LOGGER.mkdirs();
        DIRECTORY_SYNCLET_INPUT.mkdirs();
        DIRECTORY_SYNCLET_OUTPUT.mkdirs();
    }

    /**
     * Returns the singleton instance. It created the instance the first time it
     * is called.
     *
     * @return the singleton instance.
     */
    public static synchronized Configuration getConfiguration() {

        if (singleton == null) {
            singleton = new Configuration();

            try {

                URL               configPath = ConfigTools.fixURL(getConfigPath());
                ConfigClassLoader cl         =
                    new ConfigClassLoader(new URL[] {configPath},
                                          Configuration.class.getClassLoader());

                singleton.setClassLoader(cl);

                singleton.configureLogging();

                singleton.configurePersistentStoreManager();

                singleton.configurePipelineManager();

            } catch (Exception e) {

                String msg = "Fatal error creating the configuration object ";
                log.fatal(msg, e);
            }
            singleton.enableDirectoryMonitor();
        }

        return singleton;
    }

    // ---------------------------------------------------------- Public methods

    /**
     * Getter for property serverConfig.
     * @return Value of property serverConfig.
     */
    public ServerConfiguration getServerConfig() {

        return (ServerConfiguration)ConfigTools.getBeanInstance(classLoader,
                                                                BEAN_SERVER_CONFIGURATION);
    }

    /**
     * Creates and returns a new SyncStrategy instance from the configured
     * <i>singleton.getServerConfig().getEngineConfiguration().strategy</i> property
     *
     * @return the newly created instance
     */
    public SyncStrategy getStrategy() {
        try {
            return (SyncStrategy)ConfigTools.getNewBeanInstance(
                classLoader,
                singleton.getServerConfig().getEngineConfiguration().getStrategy()
            );
        } catch (Exception e) {
            log.fatal("Error creating the SyncStrategy object", e);
        }

        return null;
    }

    /**
     * Returns the configured PersistentStoreManager.
     * @return the configured PersistentStoreManager.
     */
    public PersistentStore getStore() {
        return persistentStoreManager;
    }

    /**
     * Creates and returns a new Officer instance from the configured
     * <i>singleton.getServerConfig().getEngineConfiguration().getOfficer</i> property
     *
     * @return the newly created instance
     */
    public Officer getOfficer() {
        try {
            return (Officer)ConfigTools.getBeanInstance(
                classLoader,
                singleton.getServerConfig().getEngineConfiguration().getOfficer()
            );
        } catch (Exception e) {
            log.fatal("Error creating the Officer object", e);
        }

        return null;
    }


    /**
     * Creates and returns a new SessionHandler instance from the configured
     * <i>singleton.getServerConfig().getEngineConfiguration().sessionHandler</i> property
     *
     * @return the newly created instance
     */
    public SessionHandler getSessionHandler() {
        try {
            return (SessionHandler)ConfigTools.getNewBeanInstance(
                classLoader,
                singleton.getServerConfig().getEngineConfiguration().getSessionHandler()
            );
        } catch (Exception e) {
            log.fatal("Error creating the SessionHandler object", e);
        }

        return null;
    }

    /**
     * Returns the configured PipelineManager
     *
     * @return the configured PipelineManager
     */
    public PipelineManager getPipelineManager() {
        return pipelineManager;
    }

    /**
     * Creates and returns a new UserManager instance from the configured
     * <i>singleton.getServerConfig().getEngineConfiguration().userManager</i> property
     *
     * @return the newly created instance
     */
    public UserManager getUserManager() {
        try {
            return (UserManager)ConfigTools.getNewBeanInstance(
                classLoader,
                singleton.getServerConfig().getEngineConfiguration().getUserManager()
            );
        } catch (Exception e) {
            log.fatal("Error creating the UserManager object", e);
        }

        return null;
    }


    /**
     * Creates and returns a new PipelineManager instance from the configured
     * <i>serverConfig.getEngineConfiguration().pipelineManager</i> property
     *
     * @return the newly created instance
     */
    public DeviceInventory getDeviceInventory() {
        try {
            return (DeviceInventory)ConfigTools.getBeanInstance(
                classLoader,
                singleton.getServerConfig().getEngineConfiguration().getDeviceInventory()
            );
        } catch (Exception e) {
            log.fatal("Error creating the DeviceInventory object", e);
        }

        return null;
    }

    /**
     * Creates and returns a new PipelineManager instance from the configured
     * <i>serverConfig.getEngineConfiguration().pipelineManager</i> property
     *
     * @return the newly created instance
     */
    public DataTransformerManager getDataTransformerManager() {
        String dataTransformerManager =
            singleton.getServerConfig().getEngineConfiguration().getDataTransformerManager();

        if (dataTransformerManager == null || dataTransformerManager.equals("")) {
            if (log.isTraceEnabled()) {
                log.trace("DataTransformerManager not set");
            }
            return null;
        }

        try {
            return (DataTransformerManager)ConfigTools.getBeanInstance(
                classLoader,
                dataTransformerManager
            );
        } catch (Exception e) {
            log.fatal("Error creating the DataTransformerManager object", e);
        }

        return null;
    }

    /**
     * Creates and returns an array of the available LoggerConfiguration
     *
     * @return an array of the available LoggerConfiguration
     */
    public LoggerConfiguration[] getLoggers() {

        String[] loggerBeanNames = getBeanNames(PATH_LOGGER);

        if ((loggerBeanNames == null) || (loggerBeanNames.length == 0)) {
            if (log.isTraceEnabled()) {
                log.trace("No logger bean found");
            }

            return new LoggerConfiguration[0];
        }

        LoggerConfiguration[] loggers =
            new LoggerConfiguration[loggerBeanNames.length];

        for (int i = 0; i < loggerBeanNames.length; i++) {

            try {
                loggers[i] = (LoggerConfiguration)getBeanInstanceByName(
                        loggerBeanNames[i]);
            } catch (BeanException ex) {
                continue;
            }
        }

        return loggers;
    }

    /**
     * Creates and returns a Map with the available appenders
     *
     * @return a Map with the available appenders
     */
    public Map<String, Appender> getAppenders() {

        Map<String, Appender> appendersMap = new HashMap();
        appendersMap = AppendersCache.getAppenders();
        return appendersMap;
    }

    /**
     * Sets the given appender
     * @param appender the appender to set
     * @throws ConfigurationException if an error occurs
     */
    public void setAppender(Appender appender)
    throws ConfigurationException {
        if (appender == null) {
            return ;
        }

        String beanName = PATH_APPENDER + File.separator + appender.getName() + ".xml";
        File   beanFile = new File(getConfigPath() + File.separator + beanName);

        try {

            disableDirectoryMonitor();

            ConfigTools.saveBeanInstance(beanName,
                beanFile,
                appender);

            configureLogging();

        } catch (Exception ex) {
            throw new ConfigurationException("Error saving appender", ex);
        } finally {
            enableDirectoryMonitor();
        }


        if (log.isInfoEnabled()) {
            log.info("Appender '" + appender.getName() + "' saved successfully");
        }
    }

    /**
     * Returns the class name of the management panel to configure an appender
     * with the given class name. The management class is retrieved searching an
     * appender type with the given appenderClassName
     * @return the class name of the management panel of the appender with the
     *         given class name or null if not found
     * @param appenderClassName the appender class name
     * @throws ConfigurationException if an error occurs
     */
    public String getAppenderManagementPanel(String appenderClassName)
    throws ConfigurationException {

        if (appenderClassName == null) {
            return null;
        }
        String[] appenderTypeBeanNames = getBeanNames(PATH_APPENDER_TYPE);

        if ((appenderTypeBeanNames == null) ||
            (appenderTypeBeanNames.length == 0)) {

            if (log.isTraceEnabled()) {
                log.trace("No appender type bean found");
            }

            return null;
        }

        AppenderType appenderType = null;
        for (int i = 0; i < appenderTypeBeanNames.length; i++) {

            try {
                appenderType = (AppenderType)getBeanInstanceByName(
                        appenderTypeBeanNames[i]);
            } catch (BeanException ex) {
                continue;
            }
            if (appenderClassName.equals(appenderType.getClassName())) {
                return appenderType.getManagementPanel();
            }
        }

        return null;
    }

    /**
     * Saves new server configuration
     *
     * @param serverConfig the new server configuration
     *
     * @throws ConfigurationException
     */
    public void setServerConfiguration(ServerConfiguration serverConfig)
        throws ConfigurationException {

        File f = new File(getConfigPath(), BEAN_SERVER_CONFIGURATION);
        try {

            ConfigTools.saveBeanInstance(BEAN_SERVER_CONFIGURATION, f, serverConfig);

        } catch (Exception e) {
            throw new ConfigurationException("Error saving server configuration", e);
        } finally {
            enableDirectoryMonitor();
        }
    }

    /**
     * Saves the given logger configuration
     * @param config the logger to set
     * @throws com.funambol.framework.config.ConfigurationException
     */
    public void setLoggerConfiguration(LoggerConfiguration config)
    throws ConfigurationException {

        String beanName = PATH_LOGGER + File.separator + config.getName() + ".xml";
        File   beanFile = new File(getConfigPath() + File.separator + beanName);

        try {
            disableDirectoryMonitor();

            ConfigTools.saveBeanInstance(beanName,
                beanFile,
                config);

            configureLogging();

        } catch (Exception ex) {
            throw new ConfigurationException("Error saving logger configuration", ex);
        } finally {
            enableDirectoryMonitor();
        }


        if (log.isInfoEnabled()) {
            log.info("Logger '" + config.getName() + "' saved successfully");
        }
    }

    /**
     * Configures the logging system accordingly to the
     * bean files in the logger/appender directory.
     *
     * @throws ConfigurationException in case of errors setting up the logging system
     */
    public synchronized void configureLogging()
    throws ConfigurationException {
        disableDirectoryMonitor();

        try {

            resetLog4jLogging();

            handleLoggingFile();

            configureAppenders();

            configureLoggers();

        } catch (Exception t) {
            log.error("Error setting the logging", t);
            throw new ConfigurationException("Error setting the logging", t);
        } finally {
            enableDirectoryMonitor();
        }
    }

    /**
     * Loads and instantiate a bean by its config name. In this case the bean
     * is not looked up in the internal configuration map, but is created
     * directly by. A new instance is create at each call.
     *
     * @param beanName the complete beanName
     *
     * @return the bean instance
     *
     * @throws BeanException in case of instantiation error
     */
    public Object getBeanInstanceByName(String beanName)
    throws BeanException {
        return getBeanInstanceByName(beanName, false);
    }

    /**
     * Loads and instantiate a bean by its config name. In this case the bean
     * is not looked up in the internal configuration map, but is created
     * directly by.
     *
     * If <code>cachedInstance</code> is true, this methods tries to return the
     * last instance created with the giveng server bean name. A new instance is
     * returned only when the server bean configuration has changed.
     *
     * @param beanName the complete beanName
     * @param cachedInstance should the instance be extracted from the cache?
     *
     * @return the bean instance
     *
     * @throws BeanException in case of instantiation error
     */
    public Object getBeanInstanceByName(String beanName, boolean cachedInstance)
    throws BeanException {
        if (cachedInstance) {
            return ConfigTools.getBeanInstance(classLoader, beanName);
        }

        return ConfigTools.getNewBeanInstance(classLoader, beanName);
    }

    /**
     * Saves an instantiate of a server bean to the given name under the config
     * path.
     * @param o the object to save
     * @param beanName the complete beanName
     * @throws BeanException in case of instantiation error
     */
    public void setBeanInstance(String beanName, Object o)
    throws BeanException {
        ConfigTools.saveBeanInstance(beanName, new File(getConfigPath(), beanName), o);
    }

    /**
     * Returns the Funambol home as set as follows (the first value found is
     * taken):
     * <ul>
     *  <li>the system property funambol.ds.home</li>
     *  <li>the current directory</i>
     * </ul>
     *
     * @return the Funambol home as set as discribed above.
     */
    public static String getSync4jHome() {
        return System.getProperty(PROPERTY_SYSTEM_FUNAMBOL_DS_HOME, ".");
    }

    /**
     * Returns the config path calculated as getSync4jHome() + CONFIG_PATH
     *
     * @return the config path calculated as getSync4jHome() + CONFIG_PATH
     */
    public static String getConfigPath() {

        return getSync4jHome() + CONFIG_PATH;
    }

    /**
     * Redefined to print an understandable view of the singleton object.
     *
     * @return the string representing the singleton instance
     */
    public String toString() {
        StringBuffer        sb           = new StringBuffer();

        ServerConfiguration serverConfig = singleton.getServerConfig();

        sb.append(getClass().getName())
          .append('\n')
          .append("serverInfo: ")
          .append(serverConfig.getServerInfo())
          .append('\n')
          .append("engineConfiguration: ")
          .append(serverConfig.getEngineConfiguration())
          .append('\n')
          .append("classLoader: " + classLoader)
          .append('\n');

        return sb.toString();
    }

    /**
     * Called when the DirectoryMonitor detects a change in the configuration
     * directory. See {@link com.funambol.server.tool.directorymonitor.FileChangeListener}
     *                 com.funambol.server.tool.directorymonitor.FileChangeListener
     * @param fileChangeEvent the event fired to notify a change
     */
    public void fileChange(FileChangeEvent fileChangeEvent) {
        if (fileChangeEvent == null) {
            return;
        }
        File fileChanged = fileChangeEvent.getFile();

        if (fileChanged == null) {
            log.error("Detected a change on the configuration directory but the " +
                      "changed file is null");
        }

        if (log.isTraceEnabled()) {
            StringBuilder sb = new StringBuilder("Configuration changed. ");
            if (EventType.FILE_ADDED == fileChangeEvent.getEventType()) {
                sb.append("Detected new file: " + fileChanged.getPath());
            } else if (EventType.FILE_DELETED == fileChangeEvent.getEventType()) {
                sb.append("File deleted: " + fileChanged.getPath());
            } else if (EventType.FILE_CHANGED == fileChangeEvent.getEventType()) {
                sb.append("File changed: " + fileChanged.getPath());
            } else {
                sb.append("No recognized event: " + fileChangeEvent);
            }
            log.trace(sb.toString());
        }

        if (isALoggingConfigurationFile(fileChanged)) {
            try {
                if (log.isInfoEnabled()) {
                    log.info("Logging configuration changed ('" + fileChanged.getPath() +
                        "'). Configuring logging");
                }

                configureLogging();

                if (log.isInfoEnabled()) {
                    log.info("Logging configured");
                }
            } catch (Exception ex) {
                log.error("Error configuring the logging", ex);
            }
        }

        if (isAPersistentStoreConfigurationFile(fileChanged)) {
            try {
                if (log.isInfoEnabled()) {
                    log.info("Persistent stores configuration changed ('" + fileChanged.getPath() +
                        "'). Configuring the PersistentStoreManager");
                }

                configurePersistentStoreManager();

                if (log.isInfoEnabled()) {
                    log.info("PersistentStoreManager configured");
                }
            } catch (Exception ex) {
                log.error("Error configuring the PersistentStoreManager", ex);
            }

        }

        if (isASyncletConfigurationFile(fileChanged)) {
            try {
                if (log.isInfoEnabled()) {
                    log.info("Pipeline configuration changed ('" + fileChanged.getPath() +
                        "'). Configuring the PipelineManager");
                }

                configurePipelineManager();

                if (log.isInfoEnabled()) {
                    log.info("PipelineManager configured");
                }
            } catch (Exception ex) {
                log.error("Error configuring the PipelineManager", ex);
            }
        }
    }

    /**
     * Releases the resources
     */
    public void release() {
        if (directoryMonitor != null) {
            if (log.isInfoEnabled()) {
                log.info("Stopping configuration monitor [" + directoryMonitor.toString() + "]");
            }
            directoryMonitor.stopMonitor();
            directoryMonitor = null;
        }
    }

    // --------------------------------------------------------- Private methods

    /**
     * Configures the PersistentStoreManager accordingly to the
     * bean files in <code>DIRECTORY_PERSISTENT_STORE</code> directory.
     *
     * @throws ConfigurationException in case of errors setting up the persistent
     *         store manager
     */
    private synchronized void configurePersistentStoreManager()
    throws ConfigurationException {

        PersistentStore persistentStore = null;

        String storeBeanName =
                singleton.getServerConfig().getEngineConfiguration().getStore();

        File   storeBeanFile =
            new File(getConfigPath() + File.separator + storeBeanName);

        if (storeBeanFile.exists()) {
            try {
                persistentStore = (PersistentStore)ConfigTools.getBeanInstance(
                                      classLoader,
                                      storeBeanName
                                  );
            } catch (Exception e) {
                if (log.isTraceEnabled()) {
                    log.trace("Error creating the PersistentStoreManager (" +
                              e.toString() + "). A default one will be created.");
                }
            }
        }

        if (persistentStore == null) {
            persistentStore = new PersistentStoreManager();
        } else {
            if (!(persistentStore instanceof PersistentStoreManager)) {
                persistentStore = new PersistentStoreManager();
            }
        }
        this.persistentStoreManager = (PersistentStoreManager)persistentStore;

        String[] alreadyUsedStoreFileNames = persistentStoreManager.getStores();
        if (alreadyUsedStoreFileNames == null) {
            alreadyUsedStoreFileNames = new String[0];
        }
        File[]   availableStoreFiles       = DIRECTORY_PERSISTENT_STORE.listFiles();
        String   fileName                  = null;

        boolean storeAlreadyUsed = false;

        for (int i=0; i<availableStoreFiles.length; i++) {
            storeAlreadyUsed = false;
            fileName = PATH_PERSISTENT_STORE + "/" + availableStoreFiles[i].getName();
            for (int j=0; j<alreadyUsedStoreFileNames.length; j++) {
                if (alreadyUsedStoreFileNames[j].equals(fileName)) {
                    //
                    // The store is already used
                    //
                    storeAlreadyUsed = true;
                    break;
                }
            }
            if (!storeAlreadyUsed) {
                //
                // We have to add it to the PersistentStoreManager
                //
                if (log.isTraceEnabled()) {
                    log.trace("Adding '" + fileName + "' to the list of available stores");
                }
                try {
                    persistentStore = (PersistentStore) getBeanInstanceByName(fileName);
                    if (!(persistentStore instanceof PersistentStoreManager)) {
                        persistentStoreManager.addPersistentStore(persistentStore);
                    }
                } catch (Exception ex) {
                    log.error("Error creating a PersistentStore from '" + fileName + "'", ex);
                }

            }
        }
    }

    /**
     * Configures the PipelineManager accordingly to the
     * bean files in <code>DIRECTORY_SYNCLET</code> directory.
     *
     * @throws ConfigurationException in case of errors setting up the persistent
     *         store manager
     */
    private synchronized void configurePipelineManager()
    throws ConfigurationException {

        //
        // Checks if there is a PipelineManager.xml
        //
        try {
            handlePipelineManager();
        } catch (Exception e) {
            log.error("Error handling old PipelineManager configuration",
                      e);
            throw new ConfigurationException("Error handling old PipelineManager configuration",
                                             e);
        }

        pipelineManager = new PipelineManager();

        //
        // Reads the input processors from the input directory
        //
        String[] inputProcessorsBeanNames = getBeanNames(PATH_SYNCLET_INPUT);

        if (log.isTraceEnabled()) {

            if ((inputProcessorsBeanNames.length == 0)) {
                log.trace("No configured input processors");
            }
        }

        List inputProcessorsList = new ArrayList();
        Object o = null;
        for (int i = 0; i < inputProcessorsBeanNames.length; i++) {
            if (log.isTraceEnabled()) {
                log.trace("Adding input processor: " + inputProcessorsBeanNames[i]);
            }
            try {
                o = getBeanInstanceByName(inputProcessorsBeanNames[i]);
            } catch (BeanException ex) {
                log.error("Error creating input processor '" + inputProcessorsBeanNames[i] + "'", ex);
                continue;
            }
            if (o instanceof InputMessageProcessor) {
                inputProcessorsList.add((InputMessageProcessor)o);
            } else {
                if (log.isWarningEnabled()) {
                    log.warn(
                        "The directory '" + PATH_SYNCLET_INPUT + "' contains '"
                        + inputProcessorsBeanNames[i]
                        + "' that is not an InputMessageProcessor object");
                }
            }
        }
        pipelineManager.addInputProcessors(inputProcessorsList);

        //
        // Reads the output processors from the input directory
        //
        String[] outputProcessorsBeanNames = getBeanNames(PATH_SYNCLET_OUTPUT);

        if (log.isTraceEnabled()) {

            if ((outputProcessorsBeanNames.length == 0)) {
                log.trace("No configured output processors");
            }
        }

        List outputProcessorsList = new ArrayList();
        for (int i = 0; i < outputProcessorsBeanNames.length; i++) {
            if (log.isTraceEnabled()) {
                log.trace("Adding output processor: " + outputProcessorsBeanNames[i]);
            }
            try {
                o = getBeanInstanceByName(outputProcessorsBeanNames[i]);
            } catch (BeanException ex) {
                log.error("Error creating output processor '" + outputProcessorsBeanNames[i] + "'", ex);
                continue;
            }
            if (o instanceof OutputMessageProcessor) {
                outputProcessorsList.add((OutputMessageProcessor)o);
            } else {
                if (log.isWarningEnabled()) {
                    log.warn(
                        "The directory '" + PATH_SYNCLET_OUTPUT + "' contains '"
                        + outputProcessorsBeanNames[i]
                        + "' that is not an OutputMessageProcessor object");
                }
            }
        }
        pipelineManager.addOutputProcessors(outputProcessorsList);
    }

    /**
     * Is the given file a logging configuration file ?
     * @param f the file to check
     * @return Is the given file a logging configuration file ?
     */
    private boolean isALoggingConfigurationFile(File f) {
        if (f == null) {
            return false;
        }
        if (f.getParentFile().equals(DIRECTORY_LOGGING)) {
            return true;
        }
        if (f.getParentFile().equals(DIRECTORY_LOGGER)) {
            return true;
        }
        if (f.getParentFile().equals(DIRECTORY_APPENDER)) {
            return true;
        }
        return false;
    }

    /**
     * Is the given file a persistent store configuration file ?
     * @param f the file to check
     * @return Is the given file a persistent store configuration file ?
     */
    private boolean isAPersistentStoreConfigurationFile(File f) {
        if (f == null) {
            return false;
        }
        if (f.getParentFile().equals(DIRECTORY_PERSISTENT_STORE)) {
            return true;
        }
        return false;
    }

    /**
     * Is the given file a synclet ?
     * @param f the file to check
     * @return Is the given file a synclet ?
     */
    private boolean isASyncletConfigurationFile(File f) {
        if (f == null) {
            return false;
        }
        if (f.getParentFile().equals(DIRECTORY_SYNCLET_INPUT)) {
            return true;
        }
        if (f.getParentFile().equals(DIRECTORY_SYNCLET_OUTPUT)) {
            return true;
        }
        return false;
    }


    /**
     * Configures log4j logging based on the given LoggerConfiguration object
     *
     * @param conf the LoggerConfiguration object
     */
    private void configureLog4jLogger(LoggerConfiguration conf) {

        //
        // If this logger inherits from its parent, just skip logger
        // configuration.
        //
        if (conf.isInherit()) {
            return;
        }
        String loggerName = conf.getName();

        FunambolLoggerFactory.resetConfiguration(loggerName);

        FunambolLogger logger = FunambolLoggerFactory.getLogger(loggerName);

        //
        // Configuring appenders. First remove all old appenders
        //
        logger.removeAllAppenders();
        logger.setAdditivity(false);
        //
        // Converts the list of the users with Level.ALL to a Map needed to the
        // FunambolLogger
        //
        List<String> usersWithLevelALL = conf.getUsersWithLevelALL();
        logger.setUserLevels(null);
        if (usersWithLevelALL != null) {
            HashMap<String, org.apache.log4j.Level> usersLevel =
                new HashMap<String, org.apache.log4j.Level>(usersWithLevelALL.size());
            for (String user : usersWithLevelALL) {
                usersLevel.put(user, org.apache.log4j.Level.ALL);
            }
            logger.setUserLevels(usersLevel);
        }

        List<String> appenderList = conf.getAppenders();
        Appender     appender;

        if (appenderList != null) {
            for (String appenderName : appenderList) {
                appender = AppendersCache.getAppender(appenderName);

                if (appender == null) {

                    if (log.isWarningEnabled()) {
                        log.warn(
                            "Logger '" + loggerName
                            + "' is configured to use appender '" + appenderName
                            + "' but this appender is not available."
                            + "Check directory '<FUNAMBOL_HOME>/config/" + PATH_APPENDER
                            + "'");
                    }
                } else {
                    logger.addAppender(appender);
                }
            }
        }

        if (LoggerConfiguration.LEVEL_OFF.equalsIgnoreCase(conf.getLevel())) {
            logger.setLevel(org.apache.log4j.Level.OFF);
        } else if (LoggerConfiguration.LEVEL_FATAL.equalsIgnoreCase(conf.getLevel())) {
            logger.setLevel(org.apache.log4j.Level.FATAL);
        } else if (LoggerConfiguration.LEVEL_ERROR.equalsIgnoreCase(conf.getLevel())) {
            logger.setLevel(org.apache.log4j.Level.ERROR);
        } else if (LoggerConfiguration.LEVEL_WARN.equalsIgnoreCase(conf.getLevel())) {
            logger.setLevel(org.apache.log4j.Level.WARN);
        } else if (LoggerConfiguration.LEVEL_INFO.equalsIgnoreCase(conf.getLevel())) {
            logger.setLevel(org.apache.log4j.Level.INFO);
        }  else if (LoggerConfiguration.LEVEL_DEBUG.equalsIgnoreCase(conf.getLevel())) {
            logger.setLevel(org.apache.log4j.Level.DEBUG);
        } else if (LoggerConfiguration.LEVEL_TRACE.equalsIgnoreCase(conf.getLevel())) {
            logger.setLevel(org.apache.log4j.Level.TRACE);
        } else if (LoggerConfiguration.LEVEL_ALL.equalsIgnoreCase(conf.getLevel())) {
            logger.setLevel(org.apache.log4j.Level.ALL);
        }
    }

    /**
     * Returns an array of String that contains the bean name of the bean files
     * under the given directory.
     * @return an array of String that contains the bean name of the bean files
     *         under the given directory
     * @param directory the directory
     */
    private String[] getBeanNames(String directory) {

        File          dir   = new File(getConfigPath(),
                                       directory);
        File[]        files = dir.listFiles();
        if (files == null) {
            return new String[0];
        }
        String[]      ret   = new String[files.length];
        int           i     = 0;
        StringBuilder sb    = null;
        List          beanList = new ArrayList();
        for (; i < files.length; ++i) {
            if (files[i].isFile() && files[i].getName().endsWith(".xml")) {
                sb = new StringBuilder(directory);
                sb.append(File.separator);
                sb.append(files[i].getName());
                beanList.add(sb.toString());
            }
        }

        return (String[])beanList.toArray(new String[0]);
    }

    /**
     * Reads the logging configuration file and splits it in logger files (under logger directory).
     * After that, the logging configuration file is deleted (deprecated starting from v6)
     * @throws com.funambol.framework.tools.beans.BeanException if an error occurs
     */
    private void handleLoggingFile()
    throws BeanException {

        //
        // we configure Funambol logging using logging.xml file
        //
        String loggingBeanName =
            getServerConfig().getEngineConfiguration().getLoggingConfiguration();

        File   loggingFile     =
            new File(getConfigPath() + File.separator + loggingBeanName);

        if (!loggingFile.exists()) {
            return;
        }

        LoggingConfiguration lc = (LoggingConfiguration)ConfigTools.getNewBeanInstance(
                classLoader,
                loggingBeanName
        );

        if (lc == null) {
            return;
        }

        LoggerConfiguration conf;
        Iterator            it = lc.getLoggers().iterator();

        while (it.hasNext()) {
            conf = (LoggerConfiguration)it.next();
            //
            // By default, we set FunambolConsole and FunambolLogFile as appender
            //
            List a = new ArrayList();
            a.add("FunambolConsole");
            a.add("FunambolLogFile");
            conf.setAppenders(a);
            //
            // We store the conf object in the logger directory. The loggig.xml
            // file is deprecated starting from v6 but it is still read and
            // any configuration objects in that file are written in the
            // logger directory using the "logger name".xml as file name.
            // Note that if a logger "logger name".xml already exists, this is
            // not overwritten.
            // After that, the logging.xml file is deleted
            //
            String beanName = PATH_LOGGER + File.separator + conf.getName() + ".xml";
            File   beanFile = new File(getConfigPath() + File.separator + beanName);
            if (!beanFile.exists()) {
                ConfigTools.saveBeanInstance(beanName, beanFile, conf);
            }
        }

        ConfigTools.deleteBeanInstance(loggingBeanName, loggingFile);

        if (log.isTraceEnabled()) {
            log.trace("Logging configuration file deleted");
        }
    }

    /**
     * Reads the PipelineManager configuration file and splits it in input processor files
     * (under pipeline/input directory) and in output processor files
     * (under pipeline/output directory).
     * After that, the PipelineManager configuration file is deleted (deprecated starting from v6)
     */
    private void handlePipelineManager() throws BeanException {

        String pipelineManagerBeanName =
            getServerConfig().getEngineConfiguration().getPipelineManager();

        File   pipelineManagerFile     =
            new File(getConfigPath() + File.separator + pipelineManagerBeanName);

        if (!pipelineManagerFile.exists()) {
            return;
        }

        PipelineManager configuredPipelineManager = null;

        try {
            configuredPipelineManager =
                (PipelineManager)getBeanInstanceByName(pipelineManagerBeanName);
        } catch (Exception e) {
            log.error("Error creating the PipelineManager reading '" +
                      pipelineManagerBeanName + "'. That file will be ingored", e);
            return;
        }

        if (configuredPipelineManager == null) {
            return ;
        }

        //
        // We store the inputProcessor and outputProcessor objects
        // in pipeline/input and pipeline/output directories.
        // The PipelineManager.xml file is deprecated starting from v6 but
        // it is still read and:
        // - any configured inputProcessor in that file is
        //   written in the "input" directory using "0000.000x.classname.xml"
        //   as file name.
        // - any configured outputProcessor in that file is
        //   written in the "output" directory using "0000.000x.classname.xml"
        //   as file name.
        // After that, the pipelineManager.xml file is deleted
        //

        //
        // Handling InputMessageProcessors
        //
        InputMessageProcessor[] inputProcessors
            = configuredPipelineManager.getInputProcessors();

        String sCont = null;
        if (inputProcessors != null)  {
            for (int i=0; i<inputProcessors.length; i++) {
                sCont = String.valueOf(i+1);
                if (i < 9) {
                    sCont = "000" + sCont;
                } else if (i < 99) {
                    sCont = "00" + sCont;
                } else if (i < 999) {
                    sCont = "0" + sCont;
                }

                String className = inputProcessors[i].getClass().getName();
                String beanName = PATH_SYNCLET_INPUT + File.separator +
                                  "0000." + sCont + "." + className + ".xml";
                File   beanFile = new File(getConfigPath() + File.separator + beanName);

                if (!beanFile.exists()) {
                    ConfigTools.saveBeanInstance(beanName, beanFile, inputProcessors[i]);
                }

            }
        }


        //
        // Handling OutputMessageProcessors
        //
        OutputMessageProcessor[] outputProcessors
            = configuredPipelineManager.getOutputProcessors();

        sCont = null;
        if (outputProcessors != null)  {
            for (int i=0; i<outputProcessors.length; i++) {
                sCont = String.valueOf(i+1);
                if (i < 9) {
                    sCont = "000" + sCont;
                } else if (i < 99) {
                    sCont = "00" + sCont;
                } else if (i < 999) {
                    sCont = "0" + sCont;
                }

                String className = inputProcessors[i].getClass().getName();
                String beanName = PATH_SYNCLET_OUTPUT + File.separator +
                                  "0000." + sCont + "." + className + ".xml";
                File   beanFile = new File(getConfigPath() + File.separator + beanName);

                if (!beanFile.exists()) {
                    ConfigTools.saveBeanInstance(beanName, beanFile, outputProcessors[i]);
                }

            }
        }

        ConfigTools.deleteBeanInstance(pipelineManagerBeanName, pipelineManagerFile);

        if (log.isTraceEnabled()) {
            log.trace("PipelineManager configuration file deleted");
        }
    }

    /**
     * Resets log4j loggers
     */
    private void resetLog4jLogging() {
        FunambolLoggerFactory.resetConfiguration();
    }

    /**
     * Disables the directory monitor
     */
    private void disableDirectoryMonitor() {
        if (directoryMonitor.isAlive()) {
            directoryMonitor.disable();
        }
    }

    /**
     * Enables the directory monitor
     */
    private void enableDirectoryMonitor() {
        if (directoryMonitor.isAlive()) {
            directoryMonitor.reset();
            directoryMonitor.enable();
        } else {
            System.out.println("Starting configuration monitor" +
                               " [" + directoryMonitor.toString() + "] on: " +
                               getConfigPath() );
            directoryMonitor.runMonitor();
        }
    }

    /**
     * Reads the configuration bean files under 'appender' directory and return a map
     * with the loaded appenders.
     * @return a map with the loaded appenders. As key, the appender name is used
     */
    private Map<String, Appender> configureAppenders() {

        AppendersCache.clear();

        String[] appenderBeanNames = getBeanNames(PATH_APPENDER);
        if (log.isTraceEnabled()) {
             if (appenderBeanNames == null || appenderBeanNames.length == 0) {
                log.trace("No configured appenders");
            }
        }

        Map<String, Appender> appenders = new HashMap(appenderBeanNames.length);
        Appender              appender  = null;
        Object                o         = null;
        for (int i = 0; i < appenderBeanNames.length; i++) {
            if (log.isTraceEnabled()) {
                log.trace("Configuring appender: " + appenderBeanNames[i]);
            }
            try {
                o = getBeanInstanceByName(appenderBeanNames[i]);
            } catch (Exception ex) {
                //
                // We log exception that occurs configuring the logging on the
                // System.out
                //
                ex.printStackTrace(System.out);
                log.error("Error instantiating '" + appenderBeanNames[i] + "'", ex);
                continue;
            }
            if (!(o instanceof Appender)) {
                if (log.isWarningEnabled()) {
                    log.warn(
                            "The directory '" + PATH_APPENDER + "' contains '"
                            + appenderBeanNames[i]
                            + "' that is not an Appender object");
                }
                continue;
            }
            appender = (Appender)o;
            if (appender instanceof FileAppender) {
                String fileName = ((FileAppender)appender).getFile();
                File file = new File(fileName);
                if (!file.isAbsolute()) {
                    fileName = getSync4jHome() + File.separator + fileName;
                    ((FileAppender)appender).setFile(fileName);
                }
            }
            if (appender instanceof OptionHandler) {
                ((OptionHandler)appender).activateOptions();
            }
            AppendersCache.cacheAppender(appender);

            appenders.put(appender.getName(),
                          appender);
        }

        return appenders;
    }

    /**
     * Reads the configuration bean files under 'logger' directory and for each
     * file, configures a new logger.
     * @throws java.io.IOException
     */
    private void configureLoggers()
    throws IOException {

        String[] loggerBeanNames = getBeanNames(PATH_LOGGER);

        if (log.isTraceEnabled()) {

            if ((loggerBeanNames == null) || (loggerBeanNames.length == 0)) {
                log.trace("No configured loggers");
            }
        }

        Object o = null;
        for (int i = 0; i < loggerBeanNames.length; i++) {
            if (log.isTraceEnabled()) {
                log.trace("Configuring logger: " + loggerBeanNames[i]);
            }
            try {
                o = getBeanInstanceByName(loggerBeanNames[i]);
            } catch (Exception ex) {
                //
                // We log exception that occurs configuring the logging on the
                // System.out
                //
                ex.printStackTrace(System.out);
                log.error("Error creating logger '" + loggerBeanNames[i] + "'", ex);
                continue;
            }
            if (o instanceof LoggerConfiguration) {
                configureLog4jLogger((LoggerConfiguration)o);
            } else {
                if (log.isWarningEnabled()) {
                    log.warn(
                        "The directory '" + PATH_LOGGER + "' contains '"
                        + loggerBeanNames[i]
                        + "' that is not a LoggerConfiguration object");
                }
            }
        }
    }
}
