/**
 * Copyright (C) 2003-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */

package com.funambol.server.engine;

import com.funambol.framework.engine.SyncItem;
import com.funambol.framework.engine.SyncItemKey;
import com.funambol.framework.engine.SyncItemImpl;
import com.funambol.framework.core.Item;
import com.funambol.framework.core.ComplexData;
import com.funambol.framework.core.Target;
import com.funambol.framework.core.Source;

/**
 * Helper class that contains usefull methods used to convert <i>SyncItem</i>s
 * into <i>Item</i>s and vice versa.
 *
 * @version  $Id: SyncItemHelper.java,v 1.5 2007/01/31 07:44:45 nichele Exp $
 */
public class SyncItemHelper {

    // ---------------------------------------------------------------- Costants

    // ---------------------------------------------------------- Public methods

    /**
     * Converts a generic <i>SyncItem</i> into an <i>Item</i> object.
     * @return the <i>Item</i> object
     * @param syncItem the syncitem to convert
     * @param key the key the item is known by the client agent
     * @param includeTarget <i>true</i> if the target must be included
     * @param includeSource <i>true</i> if the source must be included
     * @param includeData <i>true</i> if data must be included
     */
    public static Item toItem(String     key                  ,
                              SyncItem   syncItem             ,
                              boolean    includeTarget        ,
                              boolean    includeSource        ,
                              boolean    includeData          ) {

        ComplexData data = null;

        Target target = null;
        Source source = null;

        if (includeTarget) {
            target = new Target(key);
        }

        if (includeSource) {
            source = new Source(key);
        }

        Item item = new Item(source,
                             target,
                             null,   // meta
                             data,
                             false); //moreData


        if (includeData) {
            item.setIncompleteInfo(true);
        }

        return item;
    }

    /**
     * Create a new <i>SyncItem</i> with the content of an existing one. The
     * old SyncItem's key becomes the mappedKey of the new SyncItem.
     * @param key the key
     * @param fromSyncItem the original item
     * @return the new item
     */
    public static SyncItem newMappedSyncItem(SyncItemKey key, SyncItem fromSyncItem) {
        SyncItemImpl syncItem
            = new SyncItemImpl(fromSyncItem.getSyncSource(),
                               key.getKeyAsString(),
                               fromSyncItem.getKey().getKeyAsString(),
                               fromSyncItem.getState(),
                               fromSyncItem.getContent(),
                               fromSyncItem.getFormat(),
                               fromSyncItem.getType(),
                               fromSyncItem.getTimestamp()
              );

        if (syncItem     instanceof SyncItemImpl &&
            fromSyncItem instanceof SyncItemImpl) {

            ((SyncItemImpl)syncItem).setProperties(
               ((SyncItemImpl)fromSyncItem).getProperties()
            );

        }

        return syncItem;
    }

    /**
     * Escapes the given string replacing:
     * <ui>
     *   <li><CODE>&amp;</CODE> with <CODE>&amp;amp;</CODE></li>
     *   <li><CODE>&lt;</CODE> with <CODE>&amp;lt;</CODE></li>
     *   <li><CODE>]]&gt;</CODE> with <CODE>]]&amp;gt;</CODE></li>
     * </ui>
     * <br/>Note that if the given string is a CDATA (it starts with <CODE>&lt;![CDATA[ </CODE>
     * and ends with <CODE>]]&gt;</CODE>) no replace
     * is done
     * @param data the data to escape
     * @return the escaped data
     */
    public static String escapeData(String data) {
        if (data == null) {
            return null;
        }
        int len = data.length();
        if (len == 0) {
            return data;
        }
        //
        // CDATA is not escaped
        //
        if (data.startsWith("<![CDATA[") && data.endsWith("]]>")) {
            return data;
        }
        //
        // creating a new StringBuilder increasing the size of 10%...this is
        // just an assumption
        //
        StringBuilder sb = new StringBuilder((int)(len * 1.1));

        for (int i = 0; i < len; i++) {
            char chr = data.charAt(i);
            if (chr == '&') {
                sb.append("&amp;");
            } else if (chr == '<') {
                sb.append("&lt;");
            } else if (chr == '>' && i > 2  &&
                       data.charAt(i-1) == ']' &&
                       data.charAt(i-2) == ']') {
                sb.append("&gt;");
            } else {
                sb.append(chr);
            }
        }
        return sb.toString();
    }

}
