/**
 * Copyright (C) 2005-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.server.inventory;

import javax.sql.DataSource;

import java.util.HashMap;

import com.funambol.framework.filter.Clause;

import com.funambol.framework.server.*;
import com.funambol.framework.server.inventory.*;
import com.funambol.framework.server.store.*;

import com.funambol.framework.tools.beans.LazyInitBean;
import com.funambol.framework.tools.beans.BeanInitializationException;

import com.funambol.server.store.DevicePersistentStore;

/**
 * This the implementation open source of DeviceInventory interface.
 * @see DeviceInventory
 *
 * @version $Id: DeviceInventoryImpl.java,v 1.6 2007/01/11 11:49:24 nichele Exp $
 */
public class DeviceInventoryImpl implements DeviceInventory, LazyInitBean {

    // --------------------------------------------------------------- Constants
    public static final String CONFIG_JNDI_DATA_SOURCE_NAME = "jndi-data-source-name";

    // ------------------------------------------------------------ Private data
    protected DevicePersistentStore devicePersistentStore   = null;

    protected DataSource datasource = null;

    /**
     * Constructor without parameters
     */
    public DeviceInventoryImpl() {
    }

    /**
     * The JNDI name of the datasource to be used
     */
    protected String jndiDataSourceName = null;

    public String getJndiDataSourceName() {
        return this.jndiDataSourceName;
    }

    public void setJndiDataSourceName(String jndiDataSourceName)
    throws PersistentStoreException {
        this.jndiDataSourceName = jndiDataSourceName;
    }

    /**
     * Returns the devicePersistentStore
     * @return the devicePersistentStore
     */
    public DevicePersistentStore getDevicePersistentStore() {
        return devicePersistentStore;
    }

    /**
     * Sets the devicePersistentStore
     * @param the devicePersistentStore
     */
    public void setDevicePersistentStore(DevicePersistentStore devicePersistentStore) {
        this.devicePersistentStore = devicePersistentStore;
    }


    /**
     * Initialization.
     *
     * @throws BeanInitializationException in case of inittialization errors
     */
    public void init() throws BeanInitializationException {

        try {

            devicePersistentStore.setJndiDataSourceName(jndiDataSourceName);
            if (devicePersistentStore instanceof LazyInitBean) {
                ((LazyInitBean)devicePersistentStore).init();
            }

            //
            // Prepares the configuration map.
            //
            HashMap props = new HashMap(1);
            props.put(CONFIG_JNDI_DATA_SOURCE_NAME, jndiDataSourceName);

            devicePersistentStore.configure(props);


        } catch (PersistentStoreException e) {
            throw new BeanInitializationException(e.getMessage(), e.getCause());
        }
    }

    /**
     * This implementation return always <code>false</code>
     * @return false
     * @throws DeviceInventoryException
     */
    public boolean getCapabilities(Capabilities caps)
    throws DeviceInventoryException {
        return false;
    }

    /**
     * The open source implementation does nothing
     * @throws DeviceInventoryException
     */
    public void setCapabilities(Capabilities caps)
    throws DeviceInventoryException {
        return;
    }

    /**
     * The open source implementation does nothing
     * @param clause the clause capabilities to match
     * @return an empty array
     * @throws DeviceInventoryException
     */
    public Capabilities[] queryCapabilities(Clause clause)
    throws DeviceInventoryException {
        return new Capabilities[0];
    }

    /* The open source implementation does nothing.
     * @param clause the clause capabilities to match
     * @return always 0
     * @throws DeviceInventoryException
     */
    public int countCapabilities(Clause clause)
    throws DeviceInventoryException {
        return 0;
    }

    /**
     * Update the identifier capabilities of the given device with the
     * specify identifier.
     *
     * @param deviceId the identifier of the device
     * @param capsId   the identifier of the capabilities
     *
     * @throws DeviceInventoryException
     */
    public void setDeviceIdCaps(String deviceId, Long capsId)
    throws DeviceInventoryException {
        try {
            devicePersistentStore.store(deviceId, capsId);
        }  catch (Exception e) {
            throw new DeviceInventoryException("Error setting the idCaps for '" +
                                                deviceId + "' (" + e.getMessage() + ")", e);
        }
    }

    /**
     * Searches the device identified by device.getDeviceId() and fills
     * the given object with the info in the inventory.
     *
     * @param device the object to be filled with informations from repository
     *        ( deviceId must be set)
     * @return true if the device was found, false otherwise
     * @throws DeviceInventoryException
     */
    public boolean getDevice(Sync4jDevice device)
    throws DeviceInventoryException {
        try {
            return devicePersistentStore.read(device);
        } catch (NotFoundException e) {
            return false;
        } catch (Exception e) {
            throw new DeviceInventoryException("Error getting the device " +
                                                device + " (" + e.getMessage() + ")", e);
        }
    }

    /**
     * Searches the device identified by device.getDeviceId() and fills the
     * given object with the info in the inventory specifying whether device
     * capabilities should be retrieved as well.
     *
     * @param device the object to be filled with informations from repository
     *               (deviceId must be set)
     * @param caps specifys whether device capabilities should be retrieved or not
     * @return true if the device was found, false otherwise
     * @throws DeviceInventoryException
     */
    public boolean getDevice(Sync4jDevice device, boolean caps)
    throws DeviceInventoryException {
        boolean found = false;
        try {
            found = devicePersistentStore.read(device);
            if (found && caps) {
                getCapabilities(device.getCapabilities());
            }
        } catch (NotFoundException e) {
            return false;
        } catch (Exception e) {
            throw new DeviceInventoryException("Error getting the device " +
                                                device + " (" + e.getMessage() + ")", e);
        }

        return found;
    }

    /**
     * Updates the given device. The capabilities of the device won't chance.
     *
     * @param device Device to be updated
     * @throws DeviceInventoryException
     */
    public void setDevice(Sync4jDevice device)
    throws DeviceInventoryException {
        try {
            devicePersistentStore.store(device);
        }  catch (Exception e) {
            throw new DeviceInventoryException("Error saving the device " +
                                                device + " (" + e.getMessage() + ")", e);
        }
    }

    /**
     * Retrieves the devices that match the given clause.
     *
     * @param clause The additional where clause
     * @return Sync4jDevice[] devices found in DB
     * @throws DeviceInventoryException
     */
    public Sync4jDevice[] queryDevices(Clause clause)
    throws DeviceInventoryException {

        try {
            return (Sync4jDevice[])devicePersistentStore.read(new Sync4jDevice(), clause);
        } catch (Exception e) {
            throw new DeviceInventoryException("Error querying the device " +
                                               " (" + e.getMessage() + ")", e);
        }
    }

    /**
     * Retrieves the devices that match the given clause specifying whether
     * device capabilities should be retrieved as well.
     *
     * @param clause The additional where clause
     * @param caps   Specifys whether device capabilities should be retrieved
     * @return Sync4jDevice[] devices found in DB
     * @throws DeviceInventoryException
     */
    public Sync4jDevice[] queryDevices(Clause clause, boolean caps)
    throws DeviceInventoryException {

        try {
            Sync4jDevice[] devices = (Sync4jDevice[])
                                     devicePersistentStore.read(new Sync4jDevice(), clause);

            if (caps) {
                //
                // We have to fill the devices with the capabilities
                //
                for (int i=0; i<devices.length; i++) {
                    getCapabilities(
                        devices[i].getCapabilities()
                    );
                }
            }

            return devices;
        } catch (Exception e) {
            throw new DeviceInventoryException("Error querying the device " +
                                               " (" + e.getMessage() + ")", e);
        }

    }

    /**
     * Returns the number of the device records in the repository that match
     * the given clause filter. If clause is null or an instance of AllClause,
     * count all records.
     *
     * @param clause The additional where clause
     * @return int number of devices
     * @throws DeviceInventoryException
     */
    public int countDevices(Clause clause)
    throws DeviceInventoryException {

        try {
            return devicePersistentStore.count(new Sync4jDevice(), clause);
        } catch (Exception e) {
            throw new DeviceInventoryException("Error querying the device " +
                                               " (" + e.getMessage() + ")", e);
        }

    }

    /**
     * Removes the given device and his capabilities
     * @param device the device to remove
     * @throws DeviceInventoryException
     */
    public void removeDevice(Sync4jDevice device)
    throws DeviceInventoryException {

        if (device == null) {
            return;
        }
        Capabilities caps = device.getCapabilities();

        removeCapabilities(caps);

        try {
            devicePersistentStore.delete(device);
        } catch (PersistentStoreException e) {
            throw new DeviceInventoryException("Error removing device:" +
                                               e.getMessage(), e);
        }
    }

    /**
     * The open source implementation does nothing
     * @param capabilities the capabilities to remove
     * @throws DeviceInventoryException
     */
    public void removeCapabilities(Capabilities capabilities)
    throws DeviceInventoryException {

    }
}
