/**
 * Copyright (C) 2005-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.server.inventory;

import java.util.HashMap;

import com.funambol.framework.filter.Clause;
import com.funambol.framework.server.Capabilities;
import com.funambol.framework.server.inventory.DeviceInventoryException;
import com.funambol.framework.server.store.*;
import com.funambol.framework.tools.beans.BeanInitializationException;
import com.funambol.framework.tools.beans.LazyInitBean;

import com.funambol.server.store.CapabilitiesPersistentStore;

/**
 * This is an implementation of DeviceInventory interface that stores the
 * capabilities on DB.
 *
 * @see com.funambol.framework.server.inventory.DeviceInventory
 *
 * @version $Id: PSDeviceInventory.java,v 1.3 2007/01/11 11:49:24 nichele Exp $
 */
public class PSDeviceInventory
extends DeviceInventoryImpl implements LazyInitBean {

    // ------------------------------------------------------------ Private data
    private CapabilitiesPersistentStore capabilitiesPersistentStore = null;

    /**
     * Constructor without parameters
     */
    public PSDeviceInventory() {
    }

    /**
     * Returns the capabilitiesPersistentStore
     * @return the capabilitiesPersistentStore
     */
    public CapabilitiesPersistentStore getCapabilitiesPersistentStore() {
        return capabilitiesPersistentStore;
    }

    /**
     * Sets the capabilitiesPersistentStore
     * @param capabilitiesPersistentStore the capabilitiesPersistentStore to set
     */
    public void setCapabilitiesPersistentStore(
            CapabilitiesPersistentStore capabilitiesPersistentStore) {
        this.capabilitiesPersistentStore = capabilitiesPersistentStore;
    }

    /**
     * Initialization.
     *
     * @throws BeanInitializationException in case of inittialization errors
     */
    public void init() throws BeanInitializationException {

        try {

            devicePersistentStore.setJndiDataSourceName(jndiDataSourceName);
            if (devicePersistentStore instanceof LazyInitBean) {
                ((LazyInitBean)devicePersistentStore).init();
            }

            capabilitiesPersistentStore.setJndiDataSourceName(jndiDataSourceName);
            if (capabilitiesPersistentStore instanceof LazyInitBean) {
                ((LazyInitBean)capabilitiesPersistentStore).init();
            }

            //
            // Prepares the configuration map.
            //
            HashMap props = new HashMap(1);
            props.put(CONFIG_JNDI_DATA_SOURCE_NAME, jndiDataSourceName);

            devicePersistentStore.configure(props);
            capabilitiesPersistentStore.configure(props);

        } catch (PersistentStoreException e) {
            throw new BeanInitializationException(e.getMessage(), e.getCause());
        }
    }

    /**
     * Searches the capabilities identified:
     * - by id
     * or, if the id is null:
     * - by the set: Model, Manufacturer, Firmware version, Software version
     *              (if all these properties are not empty. Otherwise the method has to return false)
     * and fills the given object with the capabilities in the inventory.
     * Returns true if the capabilities were found, false otherwise.
     *
     * @param caps Capabilities to get
     * @return true if capabilities were found, false otherwise
     * @throws DeviceInventoryException
     */
    public boolean getCapabilities(Capabilities caps)
            throws DeviceInventoryException
    {
        boolean found = false;
        try {
            found = capabilitiesPersistentStore.read(caps);
        } catch (NotFoundException ex) {
            return false;
        } catch (Exception ex) {
            throw new DeviceInventoryException("Error getting the capabilities " +
                                                caps + " (" + ex.getMessage() + ")", ex);
        }

        return found;
    }

    /**
     * Add or update the given capabilities to the inventory.
     * If caps.getId() is null, the capabilities have to be added to the inventory.
     * If caps.getId() isn't null but in the inventory there aren't
     * the given capabilities an add will be performed.
     * If caps.getId() isn't null and in the inventory there are the given
     * capabilities an update will be performed.
     * If the capabilities are added and the concrete implementation creates
     * a new id for the capabilities, that id is caps.setId() must be called in order to update it.
     *
     * @param caps capabilities object to be saved
     * @throws DeviceInventoryException
     */
    public void setCapabilities(Capabilities caps)
        throws DeviceInventoryException {
        try {
            capabilitiesPersistentStore.store(caps);
        } catch (Exception ex) {
            throw new DeviceInventoryException("Error setting the capabilities " +
                                                caps + " (" + ex.getMessage() + ")", ex);
        }
    }

    /**
     * Retrieves the capabilities that match the given clause.
     *
     * @param clause the clause capabilities to match
     * @return array of Cpabilities that match the given clause
     * @throws DeviceInventoryException
     */
    public Capabilities[] queryCapabilities(Clause clause)
        throws DeviceInventoryException {

        try {
            return (Capabilities[])capabilitiesPersistentStore.
                read(clause);
        } catch (Exception ex) {
            throw new DeviceInventoryException("Error querying the capabilities " +
                                               "(" + ex.getMessage() + ")", ex);
        }
    }

    /**
     * Returns the number of the capabilities records in the repository that
     * match the given clause filter.
     * If clause is null or an instance of AllClause, count all records.
     *
     * @param clause the clause capabilities to match
     * @return number of capabilities that match the given clause
     * @throws DeviceInventoryException
     */
    public int countCapabilities(Clause clause)
            throws DeviceInventoryException
    {
        try {
            return capabilitiesPersistentStore.count(clause);
        }  catch (Exception ex) {
            throw new DeviceInventoryException("Error counting the capabilities " +
                                               "(" + ex.getMessage() + ")", ex);
        }

    }

    /**
     * Removes the given capabilities
     * @param capabilities the capabilities to remove
     * @throws DeviceInventoryException
     */
    public void removeCapabilities(Capabilities capabilities)
        throws DeviceInventoryException {
        if (capabilities == null) {
            return;
        }
        try {
            capabilitiesPersistentStore.delete(capabilities);
        } catch (PersistentStoreException ex) {
            throw new DeviceInventoryException("Error removing capabilities (" +
                                                ex.getMessage() + ")", ex);
        }
    }
}
