/**
 * Copyright (C) 2005-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.server.notification;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import com.funambol.framework.core.Alert;

import com.funambol.framework.filter.Clause;
import com.funambol.framework.filter.WhereClause;

import com.funambol.framework.logging.FunambolLogger;
import com.funambol.framework.logging.FunambolLoggerFactory;
import com.funambol.framework.logging.Sync4jLoggerName;
import com.funambol.framework.notification.Message;
import com.funambol.framework.notification.NotificationException;
import com.funambol.framework.notification.builder.DSNotificationMessageBuilder;
import com.funambol.framework.notification.sender.Sender;

import com.funambol.framework.security.Sync4jPrincipal;

import com.funambol.framework.server.Sync4jDevice;
import com.funambol.framework.server.inventory.DeviceInventoryException;
import com.funambol.framework.server.store.PersistentStoreException;

import com.funambol.server.config.Configuration;

/**
 * This class is an implementation of NotificationEngine interface
 *
 * @see NotificationEngine
 *
 * @version $Id: NotificationEngineImpl.java,v 1.16 2007/02/03 15:25:35 nichele Exp $
 */
public class NotificationEngineImpl extends AbstractNotificationEngine {

    // ------------------------------------------------------------ Private data
    private static final FunambolLogger log =
        FunambolLoggerFactory.getLogger(Sync4jLoggerName.SERVER_NOTIFICATION);

    // ------------------------------------------------------------- Constructor
    public NotificationEngineImpl(Configuration config) {
        super(config);
    }

    // ---------------------------------------------------------- Public Methods

    /**
     * Sends a notification message to the given device.
     *
     * @param deviceId device identifier
     * @param alerts   array of Alert command for the datastore to sync, used
     *                 to create the binary message to send
     * @param uimode   user intraction mode
     */
    public void sendNotificationMessage(String  deviceId,
                                        Alert[] alerts  ,
                                        int     uimode  )
    throws NotificationException {

        int sessionId = 0;
        DSNotificationMessageBuilder builder = null;
        Sender sender = null;

        if (log.isTraceEnabled()) {
            log.trace("Try to notify device '" + deviceId + "'");
        }

        Sync4jDevice device = new Sync4jDevice(deviceId);
        boolean deviceFound = false;
        try {
            deviceFound = readDeviceWithoutCapabilities(device);
        } catch(DeviceInventoryException e) {
            log.error("Error reading the device '" + deviceId +
                      "'", e);
            throw new NotificationException("Error reading device '" + deviceId + "'");
        }
        if (!deviceFound) {
            if (log.isTraceEnabled()) {
                log.trace("Device '" + deviceId + "' not found");
            }
            throw new NotificationException("Device '" + deviceId + "' not found");
        }

        String notificationBuilder = device.getNotificationBuilder();
        String notificationSender  = device.getNotificationSender();

        if (notificationBuilder == null || "".equals(notificationBuilder)) {
            if (log.isTraceEnabled()) {
                log.trace("No notification builder configured for '" + deviceId +
                           "'");
            }
            return;
        }

        if (notificationSender == null || "".equals(notificationSender)) {
            if (log.isTraceEnabled()) {
                log.trace("No notification sender configured for '" + deviceId +
                           "'");
            }
            return;
        }

        try {
            builder = (DSNotificationMessageBuilder)
                getConfiguration().getBeanInstanceByName(notificationBuilder);
        } catch (Exception e) {
            log.error("Error creating the builder '" + notificationBuilder +
                      "'", e);

            throw new NotificationException(e.getMessage());
        }

        try {
            sender = (Sender)
                getConfiguration().getBeanInstanceByName(notificationSender);
        } catch (Exception e) {
            log.error("Error creating the sender '" + notificationBuilder +
                      "'", e);
            throw new NotificationException(e.getMessage());
        }

        try {
            String serverId =
                getConfiguration().getServerConfig().getServerInfo().getDevID();
            Float protocolVersion = new Float(1.2);

            //build message
            Message message = null;
            String deviceAddress = device.getAddress();
            String serverPWD     = device.getServerPassword();
            byte[] serverNonce   = device.getServerNonce();
            try {
                message = builder.buildMessage(sessionId                   ,
                                               deviceId                    ,
                                               deviceAddress               ,
                                               serverId                    ,
                                               serverPWD                   ,
                                               serverNonce                 ,
                                               alerts                      ,
                                               uimode                      ,
                                               protocolVersion.floatValue());
            } catch (java.lang.AbstractMethodError err) {
                if (log.isWarningEnabled()) {
                    log.warn("The configured builder (" +
                             notificationBuilder + ") doesn't seem a valid builder." +
                             " Trying to use the v3 interface");
                }
                //
                // Using reflection to handle previous officer version (v3)
                //
                message = handleOldBuilderInstance(builder,
                                                  sessionId,
                                                  deviceAddress,
                                                  serverId,
                                                  serverPWD,
                                                  serverNonce,
                                                  alerts,
                                                  uimode,
                                                  protocolVersion.floatValue());
            }

            //send message
            try {
                sender.sendNotificationMessage(device, message);
            } catch (java.lang.AbstractMethodError err) {
                if (log.isWarningEnabled()) {
                    log.warn("The configured sender (" +
                             notificationSender + ") doesn't seem a valid sender." +
                             " Trying to use the v3 interface");
                }

                //
                // Using reflection to handle previous officer version (v3)
                //
                handleOldSenderInstance(sender, device, message);
            }
        } catch (Throwable e) {
            log.error("Error sending message: " + e.toString());
            throw new NotificationException("Error sending message: " + e.getMessage());
        }
    }


    /**
     * Sends a notification message to all devices of the principals with the
     * given username.
     *
     * @param username the username
     * @param alerts   array of Alert command for the datastore to sync, used
     *                 to create the binary message to send
     * @param uimode   user intraction mode
     */
    public void sendNotificationMessages(String  username,
                                         Alert[] alerts  ,
                                         int     uimode  )
    throws NotificationException {

        Clause clause = new WhereClause("username",
                                        new String[] {username},
                                        WhereClause.OPT_EQ,
                                        true
                        );

        Sync4jPrincipal[] principals = null;
        try {
            principals = (Sync4jPrincipal[]) getStore().read(new
                    Sync4jPrincipal(), clause);
        } catch (PersistentStoreException e) {
            log.error("Error reading principals", e);
            throw new NotificationException("Error reading principals: " + e.getMessage());
        }

         if (principals == null || principals.length == 0) {
             if (log.isTraceEnabled()) {
                 log.trace("No principal found with username '" + username + "'");
                 log.trace("No devices to notify");
             }
         }

         int num = principals.length;

         for (int i=0; i<num; i++) {
            try {
                sendNotificationMessage(principals[i].getDeviceId(), alerts,
                                        uimode);
            } catch (Throwable ex) {
                //
                // Ignore it...we have to notify next device...
                //
            }
         }
    }

    /**
     * Calls, using reflection, the sendNotificationMessage according to the previous sender interface
     * (the one in v3) that declared sendNotificationMessage(Sync4jDevice, byte[])
     * @param sender the sender
     * @param device the device
     * @param message the message to send
     */
    private void handleOldSenderInstance(Sender sender, Sync4jDevice device, Message message)
    throws NotificationException {
        //
        // Using reflection to handle the old implementations
        // of the Sender
        //
        byte[] messageContent = message.getMessageContent();

        Method  method         = null;
        try {
            method = sender.getClass().getMethod("sendNotificationMessage",
                                                  new Class[] {Sync4jDevice.class, byte[].class}
                     );

            method.invoke(sender, new Object[] {device, messageContent});

        } catch (InvocationTargetException e) {
            Throwable target = e.getTargetException();
            if (target instanceof NotificationException) {
               throw (NotificationException)target;
            }

            throw new NotificationException(
                "Error invoking 'sendNotificationMessage' method", target
            );
        } catch (Throwable e) {
            throw new NotificationException(
                "Error invoking 'sendNotificationMessage' method", e
            );

        }
    }

    /**
     * Calls, using reflection, the buildMessage according to the previous builder interface
     * (the one in v3) that declared:
     * <p><code>
     *    byte[] buildMessage(int     sessionId      ,
     *                        String  deviceAddress  ,
     *                        String  serverId       ,
     *                        String  serverPw       ,
     *                        byte[]  serverNonce    ,
     *                        Alert[] alerts         ,
     *                        int     uimode         ,
     *                        float   protocolVersion)
     * </code>
     *
     */
    private Message handleOldBuilderInstance(DSNotificationMessageBuilder builder,
                                             int     sessionId      ,
                                             String  deviceAddress  ,
                                             String  serverId       ,
                                             String  serverPw       ,
                                             byte[]  serverNonce    ,
                                             Alert[] alerts         ,
                                             int     uimode         ,
                                             float   protocolVersion)

    throws NotificationException {
        //
        // Using reflection to handle the old implementations
        // of the Builder
        //
        Method  method         = null;
        byte[]    messageContent = null;
        try {
            method = builder.getClass().getMethod("buildMessage",
                                                  new Class[] {
                                                      Integer.TYPE,
                                                      String.class,
                                                      String.class,
                                                      String.class,
                                                      byte[].class,
                                                      Alert[].class,
                                                      Integer.TYPE,
                                                      Float.TYPE
                                                  }
                     );

            messageContent = (byte[])method.invoke(builder,
                                                   new Object[] {
                                                       Integer.valueOf(sessionId),
                                                       deviceAddress,
                                                       serverId,
                                                       serverPw,
                                                       serverNonce,
                                                       alerts,
                                                       Integer.valueOf(uimode),
                                                       new Float(protocolVersion)
                                                   }
                               );

        } catch (InvocationTargetException e) {
            Throwable target = e.getTargetException();
            if (target instanceof NotificationException) {
               throw (NotificationException)target;
            }

            throw new NotificationException(
                "Error invoking 'buildMessage' method", target
            );
        } catch (Throwable e) {
            throw new NotificationException(
                "Error invoking 'buildMessage' method", e
            );

        }
        return new Message(Message.Type.STANDARD_1_2_NOTIFICATION_MESSAGE_TYPE,
                           messageContent);
    }

}
