/**
 * Copyright (C) 2006-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */

package com.funambol.server.tools.directorymonitor;

import java.io.File;

import java.util.*;

import com.funambol.server.tools.directorymonitor.FileChangeEvent.EventType;

/**
 * It's a director  monitor able to monitor a directory and its sub directories.
 * An <code>IllegalArgumentException</code> is thrown if the monitor is created
 * with a file that doesn't exist or if is not a directory.
 * <br/>
 * It's a Thread that every <code>scanPeriod</code> checks the directory and its
 * sub directories in order to detect new/updated/deleted files (not directory) and
 * fires a <code>FileChangeEvent</code> for any changes. Note that jsut the files
 * are monitored.
 * It notifies the registered <code>FileChangeListeners</code> when a modification
 * on the monitored directory occurs.
 *
 * @version $Id: DirectoryMonitor.java,v 1.7 2007/02/07 11:40:11 nichele Exp $
 */
public class DirectoryMonitor extends Thread {

    // --------------------------------------------------------------- Constants
    private static final String THREAD_NAME = "funambol-directory-monitor";

    // ------------------------------------------------------------ Private data

    /** List of FileChangeListener */
    private List<FileChangeListener> fileChangeListeners = null;

    /** The period between two scans */
    private long scanPeriod;

    /** Is it the monitoring enabled */
    private boolean enabled = true;

    /** The directory to monitor */
    private File directory = null;

    /** The contained files */
    private Map<File, Long> containedFiles = null;

    // -------------------------------------------------------------- Properties


    // ------------------------------------------------------------- Constructor

    /**
     * Creates a new FileSystemMonitor with the given scanPeriod
     * @param directory the directory to monitor
     * @param scanPeriod the scan period
     */
    public DirectoryMonitor(File directory, long scanPeriod) {

        if (directory == null || !directory.isDirectory()) {
            throw new IllegalArgumentException("The given file is not a directory");
        }

        this.setPriority(Thread.MIN_PRIORITY);
        this.setName(THREAD_NAME + "-" + Integer.toString(this.hashCode(), 16));
        this.fileChangeListeners = new ArrayList();

        this.scanPeriod = scanPeriod;
        this.directory  = directory;
    }

    /**
     * Creates a new FileSystemMonitor with the given scanPeriod
     * @param directory the directory to monitor
     * @param scanPeriod the scan period
     */
    public DirectoryMonitor(String directory, long scanPeriod) {

        this(new File(directory), scanPeriod);
    }


    // ---------------------------------------------------------- Public Methods

    /**
     * Is the monitor enabled ?
     * @return <CODE>true</CODE> if the monitor is enabled, <CODE>false</CODE> otherwise
     */
    public boolean isEnabled() {
        return enabled;
    }

    /**
     * Runs the monitor loading the current state of the monitored files.
     * If the monitor is already alive, it's just enabled (if disabled).
     */
    public void runMonitor() {
        if (!this.isAlive()) {
            this.containedFiles = getContainedFiles(this.directory);
            this.start();
        }
        if (!this.enabled) {
            this.enabled = true;
        }
    }

    /**
     * Disables the monitoring
     */
    public void disable() {
        this.enabled = false;
    }

    /**
     * Enables the monitoring
     */
    public void enable() {
        this.enabled = true;
    }

    /**
     * Resets the monitor reloading the state of the monitored directory.
     */
    public void reset() {
        this.containedFiles = getContainedFiles(this.directory);
    }

    /**
     * Stops the monitor
     */
    public void stopMonitor() {
        if (this.isAlive()) {
            this.interrupt();
        }
    }

    /**
     * Adds a new <CODE>FileChangeListener</CODE>. If the listener is already registered,
     * nothing is done.
     * @param listener the <CODE>FileChangeListener</CODE> to add
     */
    public void addFileChangeListener(FileChangeListener listener) {
        if (listener == null) {
            return ;
        }
        synchronized (fileChangeListeners) {
            if (fileChangeListeners.indexOf(listener) == -1) {
                fileChangeListeners.add(listener);
            }
        }
    }

    /**
     * Removes the given <CODE>FileChangeListener<CODE>
     * @param listener the <CODE>FileChangeListener</CODE> to remove
     */
    public void removeFileChangeListener(FileChangeListener listener) {
        if (listener == null) {
            return ;
        }
        synchronized (fileChangeListeners) {
            fileChangeListeners.remove(listener);
        }
    }

    /**
     * Starts the monitoring
     */
    public void run() {

        try {
            while (true) {
                if (enabled) {
                    checkFiles();
                }
                Thread.currentThread().sleep(scanPeriod);
            }
        } catch (InterruptedException ex) {
            //
            // nothing to do
            //
        }
    }


    /**
     * Returns a string representation of this directory monitor
     * @return a string representation of this directory monitor
     */
    public String toString() {
        return this.getName();
    }
    // --------------------------------------------------------- Private methods

    /**
     * Returns a map with the contained files in the given directory and the
     * respective last modified timestamp
     * @param directory the directory
     * @return a Map containing all files and respective last modified timestamp.
     *         This map contains just the last modified timestamp for the file
     *         that are not directory. Also the subdirectory are scanned and the
     *         content is added to the Map. If the given file is not a directory
     *         a null map is returned.
     */
    private Map<File, Long> getContainedFiles(File directory) {
        if (directory == null || !directory.isDirectory()) {
            return null;
        }

        File[] files = directory.listFiles();
        Map containedFiles = new HashMap(files.length);

        Long lastModified = null;
        for (File file : files) {

           if (file.isDirectory()) {
               containedFiles.putAll(getContainedFiles(file));
               continue;
           }
           lastModified = new Long(file.lastModified());
           containedFiles.put(file, lastModified);

        }
        return containedFiles;
    }

    /**
     * Checks if the state of the file in the given map is changes.
     * @param files the files to check
     */
    private void checkFiles() {

        Map<File, Long> currentContainedFiles = null;

        List<FileChangeEvent> events = new ArrayList();

        FileChangeEvent fileChangeEvent = null;

        long scanTimeStamp = System.currentTimeMillis();

        currentContainedFiles = getContainedFiles(directory);

        if (currentContainedFiles == null && containedFiles == null) {

            //
            // This happens only is the monitored file is not a directory
            //

        } else if (currentContainedFiles != null && containedFiles != null) {

            Long lastScanlastModified = null;
            Long currentLastModified = null;

            for (File file : currentContainedFiles.keySet()) {
                lastScanlastModified = containedFiles.get(file);
                currentLastModified = currentContainedFiles.get(file);
                if (lastScanlastModified == null) {
                    //
                    // The file is new
                    //
                    fileChangeEvent = new FileChangeEvent(scanTimeStamp,
                                                          file,
                                                          EventType.FILE_ADDED);
                    events.add(fileChangeEvent);
                    continue;
                } else if (lastScanlastModified.equals(currentLastModified)) {
                    //
                    // The file is not changed
                    //
                } else {
                    //
                    // The file is changed
                    //
                    fileChangeEvent = new FileChangeEvent(scanTimeStamp,
                                                          file,
                                                          EventType.FILE_CHANGED);
                    events.add(fileChangeEvent);
                }
            }

            //
            // Checking deleted files
            //
            for (File file : containedFiles.keySet()) {
                if (currentContainedFiles.get(file) == null) {
                    //
                    // The file has been deleted
                    //
                    fileChangeEvent = new FileChangeEvent(scanTimeStamp,
                                                          file,
                                                          EventType.FILE_DELETED);
                    events.add(fileChangeEvent);
                }
            }
        }
        containedFiles = currentContainedFiles;
        fireFileChangeEvents(events);

    }

    /**
     * Fires the given events notifying the registered listeners
     * @param events the events to fire
     */
    private void fireFileChangeEvents(List<FileChangeEvent> events) {
        synchronized (fileChangeListeners) {
            for (FileChangeEvent event : events) {
                Iterator itListeners = fileChangeListeners.iterator();
                while (itListeners.hasNext()) {
                    try {
                        ((FileChangeListener) itListeners.next()).fileChange(event);
                    } catch (Error e) {
                        //
                        // Nothing to do
                        //
                    }
                }
            }
        }
    }

}
